import django_filters
from django import forms
from django.conf import settings
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_filters.widgets import SuffixedMultiWidget

from wagtail.admin.models import popular_tags_for_model
from wagtail.admin.utils import get_user_display_name
from wagtail.admin.widgets import AdminDateInput, BooleanRadioSelect, FilteredSelect
from wagtail.coreutils import get_content_languages, get_content_type_label
from wagtail.models import Locale


class DateRangePickerWidget(SuffixedMultiWidget):
    """
    A widget allowing a start and end date to be picked.
    """

    template_name = "wagtailadmin/widgets/daterange_input.html"
    suffixes = ["from", "to"]

    def __init__(self, attrs=None):
        widgets = (
            AdminDateInput(attrs={"placeholder": _("Date from")}),
            AdminDateInput(attrs={"placeholder": _("Date to")}),
        )
        super().__init__(widgets, attrs)

    def decompress(self, value):
        if value:
            return [value.start, value.stop]
        return [None, None]


class FilteredModelChoiceIterator(django_filters.fields.ModelChoiceIterator):
    """
    A variant of Django's ModelChoiceIterator that, instead of yielding (value, label) tuples,
    returns (value, label, filter_value) so that FilteredSelect can drop filter_value into
    the data-filter-value attribute.
    """

    def choice(self, obj):
        return (
            self.field.prepare_value(obj),
            self.field.label_from_instance(obj),
            self.field.get_filter_value(obj),
        )


class FilteredModelChoiceField(django_filters.fields.ModelChoiceField):
    """
    A ModelChoiceField that uses FilteredSelect to dynamically show/hide options based on another
    ModelChoiceField of related objects; an option will be shown whenever the selected related
    object is present in the result of filter_accessor for that option.

    filter_field - the HTML `id` of the related ModelChoiceField
    filter_accessor - either the name of a relation, property or method on the model instance which
        returns a queryset of related objects, or a function which accepts the model instance and
        returns such a queryset.
    """

    widget = FilteredSelect
    iterator = FilteredModelChoiceIterator

    def __init__(self, *args, **kwargs):
        self.filter_accessor = kwargs.pop("filter_accessor")
        filter_field = kwargs.pop("filter_field")
        super().__init__(*args, **kwargs)
        self.widget.filter_field = filter_field

    def get_filter_value(self, obj):
        # Use filter_accessor to obtain a queryset of related objects
        if callable(self.filter_accessor):
            queryset = self.filter_accessor(obj)
        else:
            # treat filter_accessor as a method/property name of obj
            queryset = getattr(obj, self.filter_accessor)
            if isinstance(queryset, models.Manager):
                queryset = queryset.all()
            elif callable(queryset):
                queryset = queryset()

        # Turn this queryset into a list of IDs that will become the 'data-filter-value' used to
        # filter this listing
        return queryset.values_list("pk", flat=True)


class FilteredModelChoiceFilter(django_filters.ModelChoiceFilter):
    field_class = FilteredModelChoiceField


class LocaleFilter(django_filters.ChoiceFilter):
    def filter(self, qs, language_code):
        if language_code:
            try:
                locale_id = (
                    Locale.objects.filter(language_code=language_code)
                    .values_list("pk", flat=True)
                    .get()
                )
            except Locale.DoesNotExist:
                return qs.none()
            return qs.filter(locale_id=locale_id)
        return qs


class WagtailFilterSet(django_filters.FilterSet):
    def __init__(self, data=None, queryset=None, *, request=None, prefix=None):
        super().__init__(data, queryset, request=request, prefix=prefix)

        if getattr(settings, "WAGTAIL_I18N_ENABLED", False):
            self._add_locale_filter()

    def _add_locale_filter(self):
        # Add a locale filter if the model is translatable
        # and there isn't one already.
        from wagtail.models.i18n import Locale, TranslatableMixin

        if (
            self._meta.model
            and issubclass(self._meta.model, TranslatableMixin)
            and "locale" not in self.filters
        ):
            # Only add the locale filter if there are multiple content languages
            # in the settings and the corresponding Locales exist.
            languages = get_content_languages()
            locales = set(Locale.objects.values_list("language_code", flat=True))
            choices = [(k, v) for k, v in languages.items() if k in locales]
            if len(choices) <= 1:
                return

            self.filters["locale"] = LocaleFilter(
                label=_("Locale"),
                choices=choices,
                empty_label=None,
                null_label=_("All"),
                null_value=None,
                widget=forms.RadioSelect,
            )

    @classmethod
    def filter_for_lookup(cls, field, lookup_type):
        filter_class, params = super().filter_for_lookup(field, lookup_type)

        if filter_class == django_filters.ChoiceFilter:
            params.setdefault("widget", forms.RadioSelect)
            params.setdefault("empty_label", _("All"))

        elif filter_class in [django_filters.DateFilter, django_filters.DateTimeFilter]:
            params.setdefault("widget", AdminDateInput)

        elif filter_class == django_filters.DateFromToRangeFilter:
            params.setdefault("widget", DateRangePickerWidget)

        elif filter_class == django_filters.BooleanFilter:
            params.setdefault("widget", BooleanRadioSelect)

        return filter_class, params


class ContentTypeModelChoiceField(django_filters.fields.ModelChoiceField):
    """
    Custom ModelChoiceField for ContentType, to show the model verbose name as the label rather
    than the default 'wagtailcore | page' representation of a ContentType
    """

    def label_from_instance(self, obj):
        return get_content_type_label(obj)


class ContentTypeFilter(django_filters.ModelChoiceFilter):
    field_class = ContentTypeModelChoiceField


class ContentTypeModelMultipleChoiceField(
    django_filters.fields.ModelMultipleChoiceField
):
    """
    Custom ModelMultipleChoiceField for ContentType, to show the model verbose name as the label rather
    than the default 'wagtailcore | page' representation of a ContentType
    """

    def label_from_instance(self, obj):
        return get_content_type_label(obj)


class MultipleContentTypeFilter(django_filters.ModelMultipleChoiceFilter):
    field_class = ContentTypeModelMultipleChoiceField


class UserModelMultipleChoiceField(django_filters.fields.ModelMultipleChoiceField):
    """
    Custom ModelMultipleChoiceField for user models, to show the result of
    get_user_display_name as the label rather than the default string representation
    """

    def label_from_instance(self, obj):
        return get_user_display_name(obj)


class MultipleUserFilter(django_filters.ModelMultipleChoiceFilter):
    field_class = UserModelMultipleChoiceField


class CollectionChoiceIterator(django_filters.fields.ModelChoiceIterator):
    @cached_property
    def min_depth(self):
        return self.queryset.get_min_depth()

    def choice(self, obj):
        return (obj.pk, obj.get_indented_name(self.min_depth, html=True))


class CollectionChoiceField(django_filters.fields.ModelChoiceField):
    iterator = CollectionChoiceIterator


class CollectionFilter(django_filters.ModelChoiceFilter):
    field_class = CollectionChoiceField


class RelatedFilterMixin:
    # Workaround for https://github.com/wagtail/wagtail/issues/6616 by changing
    # a filter on a related field into a filter on the primary key of instances
    # that match the filter value.

    def __init__(self, *args, use_subquery=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_subquery = use_subquery

    def filter(self, qs, value):
        filtered = super().filter(qs, value)
        if not self.use_subquery or not value:
            return filtered
        pks = list(filtered.values_list("pk", flat=True))
        return qs.filter(pk__in=pks)


class PopularTagsFilter(RelatedFilterMixin, django_filters.MultipleChoiceFilter):
    # This uses a MultipleChoiceFilter instead of a ModelMultipleChoiceFilter
    # because the queryset has been sliced, which means ModelMultipleChoiceFilter
    # cannot do further queries to validate the selected tags.
    pass


class BaseMediaFilterSet(WagtailFilterSet):
    permission_policy = None

    def __init__(
        self, data=None, queryset=None, *, request=None, prefix=None, is_searching=None
    ):
        super().__init__(data, queryset, request=request, prefix=prefix)
        collections_qs = self.permission_policy.collections_user_has_any_permission_for(
            request.user, ["add", "change"]
        )
        # Add collection filter only if there are multiple collections
        if collections_qs.count() > 1:
            self.filters["collection_id"] = CollectionFilter(
                field_name="collection_id",
                label=_("Collection"),
                queryset=collections_qs,
            )

        popular_tags = popular_tags_for_model(self._meta.model)

        if popular_tags:
            self.filters["tag"] = PopularTagsFilter(
                label=_("Tag"),
                field_name="tags__name",
                choices=[(tag.name, tag.name) for tag in popular_tags],
                widget=forms.CheckboxSelectMultiple,
                use_subquery=is_searching,
                help_text=_("Filter by up to ten most popular tags."),
            )



from django.core.exceptions import ValidationError as DjangoValidationError
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.serializers import Serializer

from wagtail.actions.convert_alias import ConvertAliasPageAction, ConvertAliasPageError
from wagtail.api.v2.utils import BadRequestError

from .base import APIAction


class ConvertAliasPageAPIAction(APIAction):
    serializer = Serializer

    def _action_from_data(self, instance, data):
        return ConvertAliasPageAction(instance, user=self.request.user)

    def execute(self, instance, data):
        action = self._action_from_data(instance, data)

        try:
            new_page = action.execute()
        except DjangoValidationError as e:
            raise ValidationError(e.message_dict)
        except ConvertAliasPageError as e:
            raise BadRequestError(e.args[0])

        serializer = self.view.get_serializer(new_page)
        return Response(serializer.data, status=status.HTTP_200_OK)



from django.test import RequestFactory, TestCase, override_settings
from django.utils.encoding import force_bytes

from wagtail.models import Site

from ..utils import (
    FieldsParameterParseError,
    get_base_url,
    parse_boolean,
    parse_fields_parameter,
)


class DynamicBaseUrl:
    def __str__(self):
        return "https://www.example.com"

    def __bytes__(self):
        return force_bytes(self.__str__())

    def decode(self, *args, **kwargs):
        return self.__bytes__().decode(*args, **kwargs)


class TestGetBaseUrl(TestCase):
    def setUp(self):
        Site.objects.all().delete()

    def prepare_site(self):
        return Site.objects.get_or_create(
            hostname="other.example.com",
            port=8080,
            root_page_id=1,
            is_default_site=True,
        )[0]

    def clear_cached_site(self, request):
        del request._wagtail_site

    def test_get_base_url_unset(self):
        self.assertIsNone(get_base_url())

    def test_get_base_url_from_request(self):
        # base url for siteless request should be None
        request = RequestFactory().get("/")
        self.assertIsNone(Site.find_for_request(request))
        self.assertIsNone(get_base_url(request))

        # base url for request with a site should be based on the site's details
        site = self.prepare_site()
        self.clear_cached_site(request)
        self.assertEqual(site, Site.find_for_request(request))
        self.assertEqual(get_base_url(request), "http://other.example.com:8080")

        # port 443 should indicate https without a port
        site.port = 443
        site.save()
        self.clear_cached_site(request)
        self.assertEqual(get_base_url(request), "https://other.example.com")

        # port 80 should indicate http without a port
        site.port = 80
        site.save()
        self.clear_cached_site(request)
        self.assertEqual(get_base_url(request), "http://other.example.com")

    @override_settings(WAGTAILAPI_BASE_URL="https://bar.example.com")
    def test_get_base_url_prefers_setting(self):
        request = RequestFactory().get("/")
        site = self.prepare_site()
        self.assertEqual(site, Site.find_for_request(request))
        self.assertEqual(get_base_url(request), "https://bar.example.com")
        with override_settings(WAGTAILAPI_BASE_URL=None):
            self.assertEqual(get_base_url(request), "http://other.example.com:8080")

    @override_settings(WAGTAILAPI_BASE_URL="https://bar.example.com")
    def test_get_base_url_from_setting_string(self):
        self.assertEqual(get_base_url(), "https://bar.example.com")

    @override_settings(WAGTAILAPI_BASE_URL=b"https://baz.example.com")
    def test_get_base_url_from_setting_bytes(self):
        self.assertEqual(get_base_url(), "https://baz.example.com")

    @override_settings(WAGTAILAPI_BASE_URL=DynamicBaseUrl())
    def test_get_base_url_from_setting_object(self):
        self.assertEqual(get_base_url(), "https://www.example.com")


class TestParseFieldsParameter(TestCase):
    # GOOD STUFF

    def test_valid_single_field(self):
        parsed = parse_fields_parameter("test")

        self.assertEqual(
            parsed,
            [
                ("test", False, None),
            ],
        )

    def test_valid_multiple_fields(self):
        parsed = parse_fields_parameter("test,another_test")

        self.assertEqual(
            parsed,
            [
                ("test", False, None),
                ("another_test", False, None),
            ],
        )

    def test_valid_negated_field(self):
        parsed = parse_fields_parameter("-test")

        self.assertEqual(
            parsed,
            [
                ("test", True, None),
            ],
        )

    def test_valid_nested_fields(self):
        parsed = parse_fields_parameter("test(foo,bar)")

        self.assertEqual(
            parsed,
            [
                (
                    "test",
                    False,
                    [
                        ("foo", False, None),
                        ("bar", False, None),
                    ],
                ),
            ],
        )

    def test_valid_star_field(self):
        parsed = parse_fields_parameter("*,-test")

        self.assertEqual(
            parsed,
            [
                ("*", False, None),
                ("test", True, None),
            ],
        )

    def test_valid_star_with_additional_field(self):
        # Note: '*,test' is not allowed but '*,test(foo)' is
        parsed = parse_fields_parameter("*,test(foo)")

        self.assertEqual(
            parsed,
            [
                ("*", False, None),
                (
                    "test",
                    False,
                    [
                        ("foo", False, None),
                    ],
                ),
            ],
        )

    def test_valid_underscore_field(self):
        parsed = parse_fields_parameter("_,test")

        self.assertEqual(
            parsed,
            [
                ("_", False, None),
                ("test", False, None),
            ],
        )

    def test_valid_field_with_underscore_in_middle(self):
        parsed = parse_fields_parameter("a_test")

        self.assertEqual(
            parsed,
            [
                ("a_test", False, None),
            ],
        )

    def test_valid_negated_field_with_underscore_in_middle(self):
        parsed = parse_fields_parameter("-a_test")

        self.assertEqual(
            parsed,
            [
                ("a_test", True, None),
            ],
        )

    def test_valid_field_with_underscore_at_beginning(self):
        parsed = parse_fields_parameter("_test")

        self.assertEqual(
            parsed,
            [
                ("_test", False, None),
            ],
        )

    def test_valid_field_with_underscore_at_end(self):
        parsed = parse_fields_parameter("test_")

        self.assertEqual(
            parsed,
            [
                ("test_", False, None),
            ],
        )

    # BAD STUFF

    def test_invalid_char(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test#")

        self.assertEqual(str(e.exception), "unexpected char '#' at position 4")

    def test_invalid_whitespace_before_identifier(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter(" test")

        self.assertEqual(str(e.exception), "unexpected whitespace at position 0")

    def test_invalid_whitespace_after_identifier(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test ")

        self.assertEqual(str(e.exception), "unexpected whitespace at position 4")

    def test_invalid_whitespace_after_comma(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test, test")

        self.assertEqual(str(e.exception), "unexpected whitespace at position 5")

    def test_invalid_whitespace_before_comma(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test ,test")

        self.assertEqual(str(e.exception), "unexpected whitespace at position 4")

    def test_invalid_unexpected_negation_operator(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test-")

        self.assertEqual(str(e.exception), "unexpected char '-' at position 4")

    def test_invalid_unexpected_open_bracket(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test,(foo)")

        self.assertEqual(str(e.exception), "unexpected char '(' at position 5")

    def test_invalid_unexpected_close_bracket(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test)")

        self.assertEqual(str(e.exception), "unexpected char ')' at position 4")

    def test_invalid_unexpected_comma_in_middle(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test,,foo")

        self.assertEqual(str(e.exception), "unexpected char ',' at position 5")

    def test_invalid_unexpected_comma_at_end(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test,foo,")

        self.assertEqual(str(e.exception), "unexpected char ',' at position 9")

    def test_invalid_unclosed_bracket(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test(foo")

        self.assertEqual(
            str(e.exception),
            "unexpected end of input (did you miss out a close bracket?)",
        )

    def test_invalid_subfields_on_negated_field(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("-test(foo)")

        self.assertEqual(str(e.exception), "unexpected char '(' at position 5")

    def test_invalid_star_field_in_wrong_position(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test,*")

        self.assertEqual(str(e.exception), "'*' must be in the first position")

    def test_invalid_negated_star(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("-*")

        self.assertEqual(str(e.exception), "'*' cannot be negated")

    def test_invalid_star_with_nesting(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("*(foo,bar)")

        self.assertEqual(str(e.exception), "unexpected char '(' at position 1")

    def test_invalid_star_with_chars_after(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("*foo")

        self.assertEqual(str(e.exception), "unexpected char 'f' at position 1")

    def test_invalid_star_with_chars_before(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("foo*")

        self.assertEqual(str(e.exception), "unexpected char '*' at position 3")

    def test_invalid_star_with_additional_field(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("*,foo")

        self.assertEqual(
            str(e.exception), "additional fields with '*' doesn't make sense"
        )

    def test_invalid_underscore_in_wrong_position(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("test,_")

        self.assertEqual(str(e.exception), "'_' must be in the first position")

    def test_invalid_negated_underscore(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("-_")

        self.assertEqual(str(e.exception), "'_' cannot be negated")

    def test_invalid_underscore_with_nesting(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("_(foo,bar)")

        self.assertEqual(str(e.exception), "unexpected char '(' at position 1")

    def test_invalid_underscore_with_negated_field(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("_,-foo")

        self.assertEqual(str(e.exception), "negated fields with '_' doesn't make sense")

    def test_invalid_star_and_underscore(self):
        with self.assertRaises(FieldsParameterParseError) as e:
            parse_fields_parameter("*,_")

        self.assertEqual(str(e.exception), "'_' must be in the first position")


class TestParseBoolean(TestCase):
    # GOOD STUFF

    def test_valid_true(self):
        parsed = parse_boolean("true")

        self.assertIs(parsed, True)

    def test_valid_false(self):
        parsed = parse_boolean("false")

        self.assertIs(parsed, False)

    def test_valid_1(self):
        parsed = parse_boolean("1")

        self.assertIs(parsed, True)

    def test_valid_0(self):
        parsed = parse_boolean("0")

        self.assertIs(parsed, False)

    # BAD STUFF

    def test_invalid(self):
        with self.assertRaises(ValueError) as e:
            parse_boolean("foo")

        self.assertEqual(str(e.exception), "expected 'true' or 'false', got 'foo'")

    def test_invalid_integer(self):
        with self.assertRaises(ValueError) as e:
            parse_boolean("2")

        self.assertEqual(str(e.exception), "expected 'true' or 'false', got '2'")



from django.db import connection
from django.db.migrations import Migration
from django.db.migrations.loader import MigrationLoader

from wagtail.blocks.migrations.migrate_operation import MigrateStreamData


class MigrationTestMixin:
    model = None
    default_operation_and_block_path = []
    app_name = None

    def init_migration(self, revisions_from=None, operations_and_block_paths=None):
        migration = Migration(
            "test_migration", "wagtail_streamfield_migration_toolkit_test"
        )
        migration_operation = MigrateStreamData(
            app_name=self.app_name,
            model_name=self.model.__name__,
            field_name="content",
            operations_and_block_paths=operations_and_block_paths
            or self.default_operation_and_block_path,
            revisions_from=revisions_from,
        )
        migration.operations = [migration_operation]

        return migration

    def apply_migration(
        self,
        revisions_from=None,
        operations_and_block_paths=None,
    ):
        migration = self.init_migration(
            revisions_from=revisions_from,
            operations_and_block_paths=operations_and_block_paths,
        )

        loader = MigrationLoader(connection=connection)
        loader.build_graph()
        project_state = loader.project_state()
        schema_editor = connection.schema_editor(atomic=migration.atomic)
        migration.apply(project_state, schema_editor)



import django_filters
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db.models import OuterRef, Subquery
from django.utils.translation import gettext_lazy as _

from wagtail.admin.filters import ContentTypeFilter, WagtailFilterSet
from wagtail.admin.widgets import AdminDateInput
from wagtail.coreutils import get_content_type_label
from wagtail.models import Page, PageLogEntry, get_page_content_types
from wagtail.permissions import page_permission_policy
from wagtail.users.utils import get_deleted_user_display_name

from .base import PageReportView


class AgingPagesReportFilterSet(WagtailFilterSet):
    last_published_at = django_filters.DateTimeFilter(
        label=_("Last published before"), lookup_expr="lte", widget=AdminDateInput
    )
    content_type = ContentTypeFilter(
        label=_("Type"),
        queryset=lambda request: get_page_content_types(include_base_page_type=False),
    )

    class Meta:
        model = Page
        fields = ["live", "last_published_at", "content_type"]


class AgingPagesView(PageReportView):
    results_template_name = "wagtailadmin/reports/aging_pages_results.html"
    page_title = _("Aging pages")
    header_icon = "time"
    filterset_class = AgingPagesReportFilterSet
    index_url_name = "wagtailadmin_reports:aging_pages"
    index_results_url_name = "wagtailadmin_reports:aging_pages_results"
    export_headings = {
        "status_string": _("Status"),
        "last_published_at": _("Last published at"),
        "last_published_by_user": _("Last published by"),
        "content_type": _("Type"),
    }
    list_export = [
        "title",
        "status_string",
        "last_published_at",
        "last_published_by_user",
        "content_type",
    ]
    any_permission_required = ["add", "change", "publish"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.user_model = get_user_model()
        self.custom_field_preprocess = self.custom_field_preprocess.copy()
        self.custom_field_preprocess["content_type"] = {
            self.FORMAT_CSV: get_content_type_label,
            self.FORMAT_XLSX: get_content_type_label,
        }

    def user_id_to_python(self, user_id):
        return self.user_model._meta.pk.to_python(user_id)

    def add_last_publisher_name_to_page(self, username_mapping, page):
        if page.last_published_by:
            # Giving the last_published_by annotation an explicit output_field type
            # causes an issue with prefetch_workflow_states when the field is a
            # ConvertedValueField. If the last user to publish the page has been
            # deleted, we will render their user id in the template, so we call
            # to_python on the value so that what's rendered matches the developer's
            # expectation in the case of complex primary key types (e.g. UUIDField).
            try:
                user_id_value = self.user_id_to_python(page.last_published_by)
            except ValidationError:
                user_id_value = page.last_published_by

            last_published_by_user = username_mapping.get(
                user_id_value, get_deleted_user_display_name(user_id=user_id_value)
            )
            page.last_published_by_user = last_published_by_user
        else:
            page.last_published_by_user = ""

    def decorate_paginated_queryset(self, queryset):
        user_ids = set(queryset.values_list("last_published_by", flat=True))

        username_mapping = {
            user.pk: user.get_username()
            for user in self.user_model.objects.filter(pk__in=user_ids)
        }
        for page in queryset:
            self.add_last_publisher_name_to_page(username_mapping, page)
        return queryset

    def get_queryset(self):
        latest_publishing_log = PageLogEntry.objects.filter(
            page=OuterRef("pk"), action__exact="wagtail.publish"
        )
        self.queryset = (
            page_permission_policy.instances_user_has_permission_for(
                self.request.user, "publish"
            )
            .exclude(last_published_at__isnull=True)
            .prefetch_workflow_states()
            .select_related("content_type")
            .annotate_approved_schedule()
            .order_by("last_published_at")
            .annotate(
                last_published_by=Subquery(latest_publishing_log.values("user")[:1])
            )
        )

        return super().get_queryset()



import logging
from functools import partial

from django.core.checks import Warning
from django.http import Http404
from django.template.response import TemplateResponse
from django.urls import URLResolver
from django.urls import path as path_func
from django.urls import re_path as re_path_func
from django.urls.resolvers import RegexPattern, RoutePattern

from wagtail.models import Page
from wagtail.url_routing import RouteResult

_creation_counter = 0

logger = logging.getLogger("wagtail.routablepage")


def _path(pattern, name=None, func=None):
    def decorator(view_func):
        global _creation_counter
        _creation_counter += 1

        # Make sure page has _routablepage_routes attribute
        if not hasattr(view_func, "_routablepage_routes"):
            view_func._routablepage_routes = []

        # Add new route to view
        view_func._routablepage_routes.append(
            (
                func(pattern, view_func, name=(name or view_func.__name__)),
                _creation_counter,
            )
        )

        return view_func

    return decorator


re_path = partial(_path, func=re_path_func)
path = partial(_path, func=path_func)

# Make route an alias of re_path for backwards compatibility.
route = re_path


class RoutablePageMixin:
    """
    This class can be mixed in to a Page model, allowing extra routes to be
    added to it.
    """

    @path("")
    def index_route(self, request, *args, **kwargs):
        request.is_preview = getattr(request, "is_preview", False)

        return TemplateResponse(
            request,
            self.get_template(request, *args, **kwargs),
            self.get_context(request, *args, **kwargs),
        )

    @classmethod
    def get_subpage_urls(cls):
        routes = []

        # Loop over this class's defined routes, in method resolution order.
        # Routes defined in the immediate class take precedence, followed by
        # immediate superclass and so on
        for klass in cls.__mro__:
            routes_for_class = []
            for val in klass.__dict__.values():
                if hasattr(val, "_routablepage_routes"):
                    routes_for_class.extend(val._routablepage_routes)

            # sort routes by _creation_counter so that ones earlier in the class definition
            # take precedence
            routes_for_class.sort(key=lambda route: route[1])

            routes.extend(route[0] for route in routes_for_class)

        return tuple(routes)

    @classmethod
    def get_resolver(cls):
        if "_routablepage_urlresolver" not in cls.__dict__:
            subpage_urls = cls.get_subpage_urls()
            cls._routablepage_urlresolver = URLResolver(
                RegexPattern(r"^/"), subpage_urls
            )

        return cls._routablepage_urlresolver

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(cls._check_path_with_regex())
        return errors

    @classmethod
    def _check_path_with_regex(cls):
        routes = cls.get_subpage_urls()
        errors = []
        for route in routes:
            if isinstance(route.pattern, RoutePattern):
                pattern = route.pattern._route
                if (
                    "(?P<" in pattern
                    or pattern.startswith("^")
                    or pattern.endswith("$")
                ):
                    errors.append(
                        Warning(
                            (
                                f"Your URL pattern {route.name or route.callback.__name__} has a "
                                "route that contains '(?P<', begins with a '^', or ends with a '$'."
                            ),
                            hint="Decorate your view with re_path if you want to use regexp.",
                            obj=cls,
                            id="wagtailroutablepage.W001",
                        )
                    )
        return errors

    def reverse_subpage(self, name, args=None, kwargs=None):
        """
        This method takes a route name/arguments and returns a URL path.
        """
        args = args or []
        kwargs = kwargs or {}

        return self.get_resolver().reverse(name, *args, **kwargs)

    def resolve_subpage(self, path):
        """
        This method takes a URL path and finds the view to call.
        """
        resolver_match = self.get_resolver().resolve(path)

        # Bind the method
        resolver_match.func = resolver_match.func.__get__(self, type(self))

        return resolver_match

    def route(self, request, path_components):
        """
        This hooks the subpage URLs into Wagtail's routing.
        """
        if self.live:
            try:
                path = "/"
                if path_components:
                    path += "/".join(path_components) + "/"

                resolver_match = self.resolve_subpage(path)
                request.routable_resolver_match = resolver_match
                view, args, kwargs = resolver_match
                return RouteResult(self, args=(view, args, kwargs))
            except Http404:
                pass

        return super().route(request, path_components)

    def serve(self, request, view=None, args=None, kwargs=None):
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        if view is None:
            return super().serve(request, *args, **kwargs)
        return view(request, *args, **kwargs)

    def render(self, request, *args, template=None, context_overrides=None, **kwargs):
        """
        This method replicates what ``Page.serve()`` usually does when ``RoutablePageMixin``
        is not used. By default, ``Page.get_template()`` is called to derive the template
        to use for rendering, and ``Page.get_context()`` is always called to gather the
        data to be included in the context.

        You can use the ``context_overrides`` keyword argument as a shortcut to override or
        add new values to the context. For example:

        .. code-block:: python

            @path('') # override the default route
            def upcoming_events(self, request):
                return self.render(request, context_overrides={
                    'title': "Current events",
                    'events': EventPage.objects.live().future(),
                })

        You can also use the ``template`` argument to specify an alternative
        template to use for rendering. For example:

        .. code-block:: python

            @path('past/')
            def past_events(self, request):
                return self.render(
                    request,
                    context_overrides={
                        'title': "Past events",
                        'events': EventPage.objects.live().past(),
                    },
                    template="events/event_index_historical.html",
                )
        """
        if template is None:
            template = self.get_template(request, *args, **kwargs)
        context = self.get_context(request, *args, **kwargs)
        context.update(context_overrides or {})
        return TemplateResponse(request, template, context)

    def serve_preview(self, request, mode_name):
        view, args, kwargs = self.resolve_subpage("/")
        return view(request, *args, **kwargs)


class RoutablePage(RoutablePageMixin, Page):
    """
    This class extends Page by adding methods which allows extra routes to be
    added to it.
    """

    class Meta:
        abstract = True



# Generated by Django 4.0.10 on 2023-09-07 12:46

from django.db import migrations, models
import wagtail.search.index


class Migration(migrations.Migration):

    dependencies = [
        ("snippetstests", "0010_delete_filterablesnippet"),
    ]

    operations = [
        migrations.CreateModel(
            name="NonAutocompleteSearchableSnippet",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("text", models.CharField(max_length=255)),
            ],
            bases=(wagtail.search.index.Indexed, models.Model),
        ),
    ]



from io import StringIO
from urllib.parse import urlsplit

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIRequest
from django.http.request import validate_host
from django.template.response import TemplateResponse
from django.utils.cache import patch_cache_control
from django.utils.translation import gettext_lazy as _


class PreviewableMixin:
    """A mixin that allows a model to have previews."""

    def make_preview_request(
        self, original_request=None, preview_mode=None, extra_request_attrs=None
    ):
        """
        Simulate a request to this object, by constructing a fake HttpRequest object that is (as far
        as possible) representative of a real request to this object's front-end URL, and invoking
        serve_preview with that request (and the given preview_mode).

        Used for previewing / moderation and any other place where we
        want to display a view of this object in the admin interface without going through the regular
        page routing logic.

        If you pass in a real request object as original_request, additional information (e.g. client IP, cookies)
        will be included in the dummy request.
        """
        dummy_meta = self._get_dummy_headers(original_request)
        request = WSGIRequest(dummy_meta)

        # Add a flag to let middleware know that this is a dummy request.
        request.is_dummy = True

        if extra_request_attrs:
            for k, v in extra_request_attrs.items():
                setattr(request, k, v)

        obj = self

        # Build a custom django.core.handlers.BaseHandler subclass that invokes serve_preview as
        # the eventual view function called at the end of the middleware chain, rather than going
        # through the URL resolver
        class Handler(BaseHandler):
            def _get_response(self, request):
                request.is_preview = True
                request.preview_mode = preview_mode
                response = obj.serve_preview(request, preview_mode)
                if hasattr(response, "render") and callable(response.render):
                    response = response.render()
                patch_cache_control(response, private=True)
                return response

        # Invoke this custom handler.
        handler = Handler()
        handler.load_middleware()
        return handler.get_response(request)

    def _get_fallback_hostname(self):
        """
        Return a hostname that can be used on preview requests when the object has no
        routable URL, or the real hostname is not valid according to ALLOWED_HOSTS.
        """
        try:
            hostname = settings.ALLOWED_HOSTS[0]
        except IndexError:
            # Django disallows empty ALLOWED_HOSTS outright when DEBUG=False, so we must
            # have DEBUG=True. In this mode Django allows localhost amongst others.
            return "localhost"

        if hostname == "*":
            # Any hostname is allowed
            return "localhost"

        # Hostnames beginning with a dot are domain wildcards such as ".example.com" -
        # these allow example.com itself, so just strip the dot
        return hostname.lstrip(".")

    def _get_dummy_headers(self, original_request=None):
        """
        Return a dict of META information to be included in a faked HttpRequest object to pass to
        serve_preview.
        """
        url = self._get_dummy_header_url(original_request)
        if url:
            url_info = urlsplit(url)
            hostname = url_info.hostname
            if not validate_host(
                hostname,
                settings.ALLOWED_HOSTS or [".localhost", "127.0.0.1", "[::1]"],
            ):
                # The hostname is not valid according to ALLOWED_HOSTS - use a fallback
                hostname = self._get_fallback_hostname()

            path = url_info.path
            port = url_info.port or (443 if url_info.scheme == "https" else 80)
            scheme = url_info.scheme
        else:
            # Cannot determine a URL to this object - cobble together an arbitrary valid one
            hostname = self._get_fallback_hostname()
            path = "/"
            port = 80
            scheme = "http"

        http_host = hostname
        if port != (443 if scheme == "https" else 80):
            http_host = f"{http_host}:{port}"
        dummy_values = {
            "REQUEST_METHOD": "GET",
            "PATH_INFO": path,
            "SERVER_NAME": hostname,
            "SERVER_PORT": port,
            "SERVER_PROTOCOL": "HTTP/1.1",
            "HTTP_HOST": http_host,
            "wsgi.version": (1, 0),
            "wsgi.input": StringIO(),
            "wsgi.errors": StringIO(),
            "wsgi.url_scheme": scheme,
            "wsgi.multithread": True,
            "wsgi.multiprocess": True,
            "wsgi.run_once": False,
        }

        # Add important values from the original request object, if it was provided.
        HEADERS_FROM_ORIGINAL_REQUEST = [
            "REMOTE_ADDR",
            "HTTP_X_FORWARDED_FOR",
            "HTTP_COOKIE",
            "HTTP_USER_AGENT",
            "HTTP_AUTHORIZATION",
            "wsgi.version",
            "wsgi.multithread",
            "wsgi.multiprocess",
            "wsgi.run_once",
        ]
        if settings.SECURE_PROXY_SSL_HEADER:
            HEADERS_FROM_ORIGINAL_REQUEST.append(settings.SECURE_PROXY_SSL_HEADER[0])
        if original_request:
            for header in HEADERS_FROM_ORIGINAL_REQUEST:
                if header in original_request.META:
                    dummy_values[header] = original_request.META[header]

        return dummy_values

    def _get_dummy_header_url(self, original_request=None):
        """
        Return the URL that _get_dummy_headers() should use to set META headers
        for the faked HttpRequest.
        """
        if url := self.full_url:
            return url
        # If no full_url is defined, try to use the original request's URL
        # to build a dummy request so that we get the correct scheme and host.
        if original_request:
            return original_request.build_absolute_uri("/")

    def get_full_url(self):
        return None

    full_url = property(get_full_url)

    DEFAULT_PREVIEW_MODES = [("", _("Default"))]
    DEFAULT_PREVIEW_SIZES = [
        {
            "name": "mobile",
            "icon": "mobile-alt",
            "device_width": 375,
            "label": _("Preview in mobile size"),
        },
        {
            "name": "tablet",
            "icon": "tablet-alt",
            "device_width": 768,
            "label": _("Preview in tablet size"),
        },
        {
            "name": "desktop",
            "icon": "desktop",
            "device_width": 1280,
            "label": _("Preview in desktop size"),
        },
    ]

    @property
    def preview_modes(self):
        """
        A list of ``(internal_name, display_name)`` tuples for the modes in which
        this object can be displayed for preview/moderation purposes. Ordinarily an object
        will only have one display mode, but subclasses can override this -
        for example, a page containing a form might have a default view of the form,
        and a post-submission 'thank you' page.
        Set to ``[]`` to completely disable previewing for this model.
        """
        return PreviewableMixin.DEFAULT_PREVIEW_MODES

    @property
    def default_preview_mode(self):
        """
        The default preview mode to use in live preview.
        This default is also used in areas that do not give the user the option of selecting a
        mode explicitly, e.g. in the moderator approval workflow.
        If ``preview_modes`` is empty, an ``IndexError`` will be raised.
        """
        return self.preview_modes[0][0]

    @property
    def preview_sizes(self):
        """
        A list of dictionaries, each representing a preview size option for this object.
        Override this property to customize the preview sizes.
        Each dictionary in the list should include the following keys:

        - ``name``: A string representing the internal name of the preview size.
        - ``icon``: A string specifying the icon's name for the preview size button.
        - ``device_width``: An integer indicating the device's width in pixels.
        - ``label``: A string for the aria label on the preview size button.

        .. code-block:: python

            @property
            def preview_sizes(self):
                return [
                    {
                        "name": "mobile",
                        "icon": "mobile-icon",
                        "device_width": 320,
                        "label": "Preview in mobile size"
                    },
                    # Add more preview size dictionaries as needed.
                ]
        """
        return PreviewableMixin.DEFAULT_PREVIEW_SIZES

    @property
    def default_preview_size(self):
        """
        The default preview size name to use in live preview.
        Defaults to ``"mobile"``, which is the first one defined in ``preview_sizes``.
        If ``preview_sizes`` is empty, an ``IndexError`` will be raised.
        """
        return self.preview_sizes[0]["name"]

    def is_previewable(self):
        """Returns ``True`` if at least one preview mode is specified in ``preview_modes``."""
        return bool(self.preview_modes)

    def serve_preview(self, request, mode_name):
        """
        Returns an HTTP response for use in object previews.

        This method can be overridden to implement custom rendering and/or
        routing logic.

        Any templates rendered during this process should use the ``request``
        object passed here - this ensures that ``request.user`` and other
        properties are set appropriately for the wagtail user bar to be
        displayed/hidden. This request will always be a GET.
        """
        return TemplateResponse(
            request,
            self.get_preview_template(request, mode_name),
            self.get_preview_context(request, mode_name),
        )

    def get_preview_context(self, request, mode_name):
        """
        Returns a context dictionary for use in templates for previewing this object.

        By default, this returns a dictionary containing the ``request`` and
        and the ``object`` itself.
        """
        return {"object": self, "request": request}

    def get_preview_template(self, request, mode_name):
        """
        Returns a template to be used when previewing this object.

        Subclasses of ``PreviewableMixin`` must override this method to return the
        template name to be used in the preview. Alternatively, subclasses can also
        override the ``serve_preview`` method to completely customise the preview
        rendering logic.
        """
        raise ImproperlyConfigured(
            "%s (subclass of PreviewableMixin) must override get_preview_template or serve_preview"
            % type(self).__name__
        )



from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404, redirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic.base import TemplateView
from django.views.generic.edit import FormMixin

from wagtail import hooks
from wagtail.actions.move_page import MovePageAction
from wagtail.admin import messages
from wagtail.admin.forms.pages import MoveForm
from wagtail.models import Page


class MoveChooseDestination(TemplateView, FormMixin):
    template_name = "wagtailadmin/pages/move_choose_destination.html"
    form_class = MoveForm

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.page_to_move = get_object_or_404(Page, id=kwargs["page_to_move_id"])
        self.page_perms = self.page_to_move.permissions_for_user(request.user)

        if not self.page_perms.can_move():
            raise PermissionDenied

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        target_parent_models = set(
            self.page_to_move.specific_class.allowed_parent_page_models()
        )
        kwargs["target_parent_models"] = target_parent_models
        kwargs["page_to_move"] = self.page_to_move

        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["page_to_move"] = self.page_to_move
        context["move_form"] = self.get_form()

        return context

    def form_valid(self, form):
        # Receive the new parent page (this should never be empty)
        if form.cleaned_data["new_parent_page"]:
            new_parent_page = form.cleaned_data["new_parent_page"]
            return redirect(
                "wagtailadmin_pages:move_confirm",
                self.page_to_move.id,
                new_parent_page.id,
            )
        return super().form_valid(form)

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        return self.form_invalid(form)


def move_confirm(request, page_to_move_id, destination_id):
    page_to_move = get_object_or_404(Page, id=page_to_move_id).specific
    # Needs .specific_deferred because the .get_admin_display_title method is called in template
    destination = get_object_or_404(Page, id=destination_id).specific_deferred

    if not Page._slug_is_available(page_to_move.slug, destination, page=page_to_move):
        messages.error(
            request,
            _(
                "The slug '%(page_slug)s' is already in use at the selected parent page. Make sure the slug is unique and try again"
            )
            % {"page_slug": page_to_move.slug},
        )
        return redirect(
            "wagtailadmin_pages:move",
            page_to_move.id,
        )

    for fn in hooks.get_hooks("before_move_page"):
        result = fn(request, page_to_move, destination)
        if hasattr(result, "status_code"):
            return result

    pages_to_move = {page_to_move}

    # The `construct_translated_pages_to_cascade_actions` hook returns translation and
    # alias pages when the action is set to "move"
    if getattr(settings, "WAGTAIL_I18N_ENABLED", False):
        for fn in hooks.get_hooks("construct_translated_pages_to_cascade_actions"):
            fn_pages = fn([page_to_move], "move")
            if fn_pages and isinstance(fn_pages, dict):
                for additional_pages in fn_pages.values():
                    pages_to_move.update(additional_pages)

    pages_to_move = list(pages_to_move)

    if request.method == "POST":
        # any invalid moves *should* be caught by the permission check in the action
        # class, so don't bother to catch InvalidMoveToDescendant
        action = MovePageAction(
            page_to_move, destination, pos="last-child", user=request.user
        )
        action.execute()

        if getattr(settings, "WAGTAIL_I18N_ENABLED", False):
            # Move translation and alias pages if they have the same parent page.
            parent_page_translations = page_to_move.get_parent().get_translations()
            for translation in pages_to_move:
                if translation.get_parent() in parent_page_translations:
                    # Move the translated or alias page to it's translated or
                    # alias "destination" page.
                    action = MovePageAction(
                        translation,
                        destination.get_translation(translation.locale),
                        pos="last-child",
                        user=request.user,
                    )
                    action.execute()

        messages.success(
            request,
            _("Page '%(page_title)s' moved.")
            % {"page_title": page_to_move.get_admin_display_title()},
            buttons=[
                messages.button(
                    reverse("wagtailadmin_pages:edit", args=(page_to_move.id,)),
                    _("Edit"),
                )
            ],
        )

        for fn in hooks.get_hooks("after_move_page"):
            result = fn(request, page_to_move)
            if hasattr(result, "status_code"):
                return result

        return redirect("wagtailadmin_explore", destination.id)

    return TemplateResponse(
        request,
        "wagtailadmin/pages/confirm_move.html",
        {
            "page_to_move": page_to_move,
            "destination": destination,
            "translations_to_move_count": len(
                [
                    translation.id
                    for translation in pages_to_move
                    if not translation.alias_of_id and translation.id != page_to_move.id
                ]
            ),
        },
    )



from typing import TYPE_CHECKING, Any, Optional

from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy, ngettext_lazy

from wagtail.admin import messages
from wagtail.admin.ui.tables import Column, TitleColumn
from wagtail.admin.views import generic
from wagtail.admin.viewsets.model import ModelViewSet
from wagtail.coreutils import get_content_languages
from wagtail.models import Locale
from wagtail.permissions import locale_permission_policy

from .forms import LocaleForm
from .utils import get_locale_usage

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponseBase


class LanguageTitleColumn(TitleColumn):
    cell_template_name = "wagtaillocales/_language_title_cell.html"

    def get_value(self, locale):
        return locale


def _can_add_locale() -> bool:
    content_languages = get_content_languages().keys()
    created_locales = Locale.objects.filter(language_code__in=content_languages)
    return created_locales.count() < len(content_languages)


class LocaleUsageColumn(Column):
    def get_value(self, locale):
        num_pages, num_others = get_locale_usage(locale)
        if num_others:
            return gettext_lazy("%(num_pages)d pages and %(num_others)d others") % {
                "num_pages": num_pages,
                "num_others": num_others,
            }

        return ngettext_lazy("%(num_pages)d page", "%(num_pages)d pages", num_pages) % {
            "num_pages": num_pages
        }


class IndexView(generic.IndexView):
    page_title = gettext_lazy("Locales")
    add_item_label = gettext_lazy("Add a locale")
    context_object_name = "locales"
    queryset = Locale.all_objects.all()
    default_ordering = "language_code"

    columns = [
        LanguageTitleColumn(
            "language",
            label=gettext_lazy("Language"),
            sort_key="language_code",
            url_name="wagtaillocales:edit",
        ),
        LocaleUsageColumn("usage", label=gettext_lazy("Usage")),
    ]

    def get_add_url(self) -> Optional[str]:
        # Removes the "Add" button from the index view.
        if not _can_add_locale():
            return None
        return super().get_add_url()


class CreateView(generic.CreateView):
    page_title = gettext_lazy("Add locale")
    success_message = gettext_lazy("Locale '%(object)s' created.")

    def dispatch(
        self, request: "HttpRequest", *args: Any, **kwargs: Any
    ) -> "HttpResponseBase":
        # Only allow access to the add view if there are locales to be added.
        if not _can_add_locale():
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)


class EditView(generic.EditView):
    success_message = gettext_lazy("Locale '%(object)s' updated.")
    error_message = gettext_lazy("The locale could not be saved due to errors.")
    context_object_name = "locale"
    queryset = Locale.all_objects.all()


class DeleteView(generic.DeleteView):
    success_message = gettext_lazy("Locale '%(object)s' deleted.")
    page_title = gettext_lazy("Delete locale")
    confirmation_message = gettext_lazy("Are you sure you want to delete this locale?")
    queryset = Locale.all_objects.all()

    def can_delete(self, locale):
        if not self.queryset.exclude(pk=locale.pk).exists():
            self.cannot_delete_message = gettext_lazy(
                "This locale cannot be deleted because there are no other locales."
            )
            return False

        if get_locale_usage(locale) != (0, 0):
            self.cannot_delete_message = gettext_lazy(
                "This locale cannot be deleted because there are pages and/or other objects using it."
            )
            return False

        return True

    def get_context_data(self, object=None):
        context = super().get_context_data()
        context["can_delete"] = self.can_delete(object)
        return context

    def form_valid(self, form):
        if self.can_delete(self.get_object()):
            return super().form_valid(form)
        else:
            messages.error(self.request, self.cannot_delete_message)
            return super().get(self.request)


class LocaleViewSet(ModelViewSet):
    icon = "site"
    model = Locale
    permission_policy = locale_permission_policy
    add_to_reference_index = False

    index_view_class = IndexView
    add_view_class = CreateView
    edit_view_class = EditView
    delete_view_class = DeleteView

    # It does not make sense to copy Locale objects as they only contain the
    # language code and are meant to be unique.
    copy_view_enabled = False

    template_prefix = "wagtaillocales/"

    def get_common_view_kwargs(self, **kwargs):
        return super().get_common_view_kwargs(
            **{
                "history_url_name": None,
                "usage_url_name": None,
                **kwargs,
            }
        )

    def get_form_class(self, for_update=False):
        return LocaleForm



# Generated by Django 4.0.7 on 2022-08-10 16:26

from django.db import migrations
import wagtail.images.models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0024_index_image_file_hash"),
    ]

    rendition_file_options = {
        "height_field": "height",
        "upload_to": wagtail.images.models.get_rendition_upload_to,
        "width_field": "width",
        "storage": wagtail.images.models.get_rendition_storage,
    }

    operations = [
        migrations.AlterField(
            model_name="image",
            name="file",
            field=wagtail.images.models.WagtailImageField(
                height_field="height",
                upload_to=wagtail.images.models.get_upload_to,
                verbose_name="file",
                width_field="width",
            ),
        ),
        migrations.AlterField(
            model_name="rendition",
            name="file",
            field=wagtail.images.models.WagtailImageField(**rendition_file_options),
        ),
    ]



import json
import unittest

from django.conf import settings
from django.contrib.auth.models import Group, Permission
from django.core import mail
from django.core.management import call_command
from django.test import TestCase, override_settings
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from taggit.models import Tag

from wagtail.admin.auth import user_has_any_page_permission
from wagtail.admin.mail import send_mail
from wagtail.admin.menu import MenuItem
from wagtail.models import Page
from wagtail.test.testapp.models import RestaurantTag
from wagtail.test.utils import WagtailTestUtils


class TestHome(WagtailTestUtils, TestCase):
    def setUp(self):
        # Login
        self.login()

    def test_simple(self):
        response = self.client.get(reverse("wagtailadmin_home"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Test Site")

    def test_accessible_skip_link_exists(self):
        response = self.client.get(reverse("wagtailadmin_home"))
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)

        # Check that the skip link is present and has the correct attributes
        skip_link = soup.select_one(".skiplink")
        self.assertIsNotNone(skip_link)
        self.assertEqual(skip_link["href"], "#main")
        self.assertEqual(skip_link["data-controller"], "w-focus")
        self.assertEqual(skip_link["data-action"], "w-focus#focus:prevent")

        # Check that the main content area has the correct ID
        main_content = soup.select_one("#main")
        self.assertIsNotNone(main_content)

    def test_admin_menu(self):
        response = self.client.get(reverse("wagtailadmin_home"))
        self.assertEqual(response.status_code, 200)
        # check that custom menu items (including classname / icon_name) are pulled in
        self.assertContains(
            response,
            '{"name": "kittens", "label": "Kittens!", "icon_name": "kitten", "classname": "kitten--test", "attrs": {"data-is-custom": "true"}, "url": "http://www.tomroyal.com/teaandkittens/"}',
        )

        # Check that the explorer menu item is here, with the right start page.
        self.assertContains(
            response,
            '[{"name": "explorer", "label": "Pages", "icon_name": "folder-open-inverse", "classname": "", "attrs": {}, "url": "/admin/pages/"}, 1]',
        )

        # There should be a link to the full-featured snippet admin in on the home page.
        self.assertContains(response, '"url": "/admin/deep/within/the/admin/"')

        # check that is_shown is respected on menu items
        response = self.client.get(reverse("wagtailadmin_home") + "?hide-kittens=true")
        self.assertNotContains(
            response,
            '{"name": "kittens", "label": "Kittens!", "icon_name": "kitten", "classname": "kitten--test", "attrs": {"data-is-custom": "true"}, "url": "http://www.tomroyal.com/teaandkittens/"}',
        )

    def test_dashboard_panels(self):
        response = self.client.get(reverse("wagtailadmin_home"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            "<p>It looks like you're making a website. Would you like some help?</p>",
        )

        # check that media attached to dashboard panels is correctly pulled in
        self.assertContains(
            response, '<script src="/static/testapp/js/clippy.js"></script>', html=True
        )

    def test_summary_items(self):
        response = self.client.get(reverse("wagtailadmin_home"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "<li>0 broken links</li>")

        # check that media attached to summary items is correctly pulled in
        self.assertContains(
            response,
            '<link href="/static/testapp/css/broken-links.css" media="all" rel="stylesheet">',
            html=True,
        )

    def test_never_cache_header(self):
        # This tests that wagtailadmins global cache settings have been applied correctly
        response = self.client.get(reverse("wagtailadmin_home"))

        self.assertIn("no-cache", response["Cache-Control"])
        self.assertIn("no-store", response["Cache-Control"])
        self.assertIn("max-age=0", response["Cache-Control"])
        self.assertIn("must-revalidate", response["Cache-Control"])

    @unittest.skipIf(
        settings.AUTH_USER_MODEL == "emailuser.EmailUser",
        "Only applicable to CustomUser",
    )
    def test_nonascii_email(self):
        # Test that non-ASCII email addresses don't break the admin; previously these would
        # cause a failure when generating Gravatar URLs
        self.create_superuser(
            username="snowman", email="☃@thenorthpole.com", password="password"
        )
        # Login
        self.assertTrue(self.client.login(username="snowman", password="password"))
        response = self.client.get(reverse("wagtailadmin_home"))
        self.assertEqual(response.status_code, 200)


class TestEditorHooks(WagtailTestUtils, TestCase):
    def setUp(self):
        self.homepage = Page.objects.get(id=2)
        self.login()

    def test_editor_css_hooks_on_add(self):
        response = self.client.get(
            reverse(
                "wagtailadmin_pages:add", args=("tests", "simplepage", self.homepage.id)
            )
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, '<link rel="stylesheet" href="/path/to/my/custom.css">'
        )

    def test_editor_js_hooks_on_add(self):
        response = self.client.get(
            reverse(
                "wagtailadmin_pages:add", args=("tests", "simplepage", self.homepage.id)
            )
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<script src="/path/to/my/custom.js"></script>')

    def test_editor_css_hooks_on_edit(self):
        response = self.client.get(
            reverse("wagtailadmin_pages:edit", args=(self.homepage.id,))
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, '<link rel="stylesheet" href="/path/to/my/custom.css">'
        )

    def test_editor_js_hooks_on_edit(self):
        response = self.client.get(
            reverse("wagtailadmin_pages:edit", args=(self.homepage.id,))
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<script src="/path/to/my/custom.js"></script>')


class TestSendMail(TestCase):
    def test_send_email(self):
        send_mail(
            "Test subject", "Test content", ["nobody@email.com"], "test@email.com"
        )

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "test@email.com")

    @override_settings(WAGTAILADMIN_NOTIFICATION_FROM_EMAIL="anothertest@email.com")
    def test_send_fallback_to_wagtailadmin_notification_from_email_setting(self):
        send_mail("Test subject", "Test content", ["nobody@email.com"])

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "anothertest@email.com")

    @override_settings(DEFAULT_FROM_EMAIL="yetanothertest@email.com")
    def test_send_fallback_to_default_from_email_setting(self):
        send_mail("Test subject", "Test content", ["nobody@email.com"])

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "yetanothertest@email.com")

    def test_send_default_from_email(self):
        send_mail("Test subject", "Test content", ["nobody@email.com"])

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "webmaster@localhost")

    def test_send_html_email(self):
        """Test that the kwarg 'html_message' works as expected on send_mail by creating 'alternatives' on the EmailMessage object"""

        send_mail(
            "Test HTML subject",
            "TEXT content",
            ["has.html@email.com"],
            html_message="<h2>Test HTML content</h2>",
        )
        send_mail("Test TEXT subject", "TEXT content", ["mr.plain.text@email.com"])

        # Check that the emails were sent
        self.assertEqual(len(mail.outbox), 2)

        # check that the first email is the HTML email
        email_message = mail.outbox[0]
        self.assertEqual(email_message.subject, "Test HTML subject")
        self.assertEqual(
            email_message.alternatives, [("<h2>Test HTML content</h2>", "text/html")]
        )
        self.assertEqual(
            email_message.body, "TEXT content"
        )  # note: plain text will always be added to body, even with alternatives
        self.assertEqual(email_message.to, ["has.html@email.com"])

        # confirm that without html_message kwarg we do not get 'alternatives'
        email_message = mail.outbox[1]
        self.assertEqual(email_message.subject, "Test TEXT subject")
        self.assertEqual(email_message.alternatives, [])
        self.assertEqual(email_message.body, "TEXT content")
        self.assertEqual(email_message.to, ["mr.plain.text@email.com"])

    def test_send_cc(self):
        send_mail(
            "Test subject",
            "Test content",
            ["nobody@email.com"],
            "test@email.com",
            cc=["cc.test@email.com"],
        )

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "test@email.com")
        self.assertEqual(mail.outbox[0].cc, ["cc.test@email.com"])

    def test_send_bcc(self):
        send_mail(
            "Test subject",
            "Test content",
            ["nobody@email.com"],
            "test@email.com",
            bcc=["bcc.test@email.com"],
        )

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "test@email.com")
        self.assertEqual(mail.outbox[0].bcc, ["bcc.test@email.com"])

    def test_send_reply_to(self):
        send_mail(
            "Test subject",
            "Test content",
            ["nobody@email.com"],
            "test@email.com",
            reply_to=["reply_to.test@email.com"],
        )

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "test@email.com")
        self.assertEqual(mail.outbox[0].reply_to, ["reply_to.test@email.com"])

    def test_send_all_extra_fields(self):
        send_mail(
            "Test subject",
            "Test content",
            ["nobody@email.com"],
            "test@email.com",
            cc=["cc.test@email.com"],
            bcc=["bcc.test@email.com"],
            reply_to=["reply_to.test@email.com"],
        )

        # Check that the email was sent
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].subject, "Test subject")
        self.assertEqual(mail.outbox[0].body, "Test content")
        self.assertEqual(mail.outbox[0].to, ["nobody@email.com"])
        self.assertEqual(mail.outbox[0].from_email, "test@email.com")
        self.assertEqual(mail.outbox[0].cc, ["cc.test@email.com"])
        self.assertEqual(mail.outbox[0].bcc, ["bcc.test@email.com"])
        self.assertEqual(mail.outbox[0].reply_to, ["reply_to.test@email.com"])


class TestTagsAutocomplete(WagtailTestUtils, TestCase):
    def setUp(self):
        self.login()
        Tag.objects.create(name="Test", slug="test")
        RestaurantTag.objects.create(name="Italian", slug="italian")
        RestaurantTag.objects.create(name="Indian", slug="indian")

    def test_tags_autocomplete(self):
        response = self.client.get(
            reverse("wagtailadmin_tag_autocomplete"), {"term": "test"}
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")
        data = json.loads(response.content.decode("utf-8"))

        self.assertEqual(data, ["Test"])

    def test_tags_autocomplete_partial_match(self):
        response = self.client.get(
            reverse("wagtailadmin_tag_autocomplete"), {"term": "te"}
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")
        data = json.loads(response.content.decode("utf-8"))

        self.assertEqual(data, ["Test"])

    def test_tags_autocomplete_different_term(self):
        response = self.client.get(
            reverse("wagtailadmin_tag_autocomplete"), {"term": "hello"}
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")
        data = json.loads(response.content.decode("utf-8"))

        self.assertEqual(data, [])

    def test_tags_autocomplete_no_term(self):
        response = self.client.get(reverse("wagtailadmin_tag_autocomplete"))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")
        data = json.loads(response.content.decode("utf-8"))
        self.assertEqual(data, [])

    def test_tags_autocomplete_custom_model(self):
        response = self.client.get(
            reverse(
                "wagtailadmin_tag_model_autocomplete", args=("tests", "restauranttag")
            ),
            {"term": "ital"},
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")
        data = json.loads(response.content.decode("utf-8"))

        self.assertEqual(data, ["Italian"])

        # should not return results from the standard Tag model
        response = self.client.get(
            reverse(
                "wagtailadmin_tag_model_autocomplete", args=("tests", "restauranttag")
            ),
            {"term": "test"},
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")
        data = json.loads(response.content.decode("utf-8"))

        self.assertEqual(data, [])

    def test_tags_autocomplete_limit(self):
        tags = [Tag(name=f"Tag {i}", slug=f"tag-{i}") for i in range(15)]
        Tag.objects.bulk_create(tags)

        # Send a request to the autocomplete endpoint with a broad search term
        response = self.client.get(
            reverse("wagtailadmin_tag_autocomplete"), {"term": "Tag"}
        )

        # Confirm the response is successful
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")

        data = json.loads(response.content.decode("utf-8"))

        # The results should be limited to avoid performance issues (#12415)
        self.assertEqual(len(data), 10)
        sorted_tags = sorted(tags, key=lambda t: t.name)
        self.assertEqual(data, [tag.name for tag in sorted_tags[:10]])


class TestMenuItem(WagtailTestUtils, TestCase):
    def setUp(self):
        self.login()
        response = self.client.get(reverse("wagtailadmin_home"))
        self.request = response.wsgi_request

    def test_menuitem_with_classname(self):
        menuitem = MenuItem(
            _("Test"),
            reverse_lazy("wagtailadmin_home"),
            classname="highlight-item",
        )
        self.assertEqual(menuitem.classname, "highlight-item")


class TestUserPassesTestPermissionDecorator(WagtailTestUtils, TestCase):
    """
    Test for custom user_passes_test permission decorators.
    testapp_bob_only_zone is a view configured to only grant access to users with a first_name of Bob
    """

    def test_user_passes_test(self):
        # create and log in as a user called Bob
        self.create_superuser(
            first_name="Bob", last_name="Mortimer", username="test", password="password"
        )
        self.login(username="test", password="password")

        response = self.client.get(reverse("testapp_bob_only_zone"))
        self.assertEqual(response.status_code, 200)

    def test_user_fails_test(self):
        # create and log in as a user not called Bob
        self.create_superuser(
            first_name="Vic", last_name="Reeves", username="test", password="password"
        )
        self.login(username="test", password="password")

        response = self.client.get(reverse("testapp_bob_only_zone"))
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_user_fails_test_ajax(self):
        # create and log in as a user not called Bob
        self.create_superuser(
            first_name="Vic", last_name="Reeves", username="test", password="password"
        )
        self.login(username="test", password="password")

        response = self.client.get(
            reverse("testapp_bob_only_zone"), HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 403)


class TestUserHasAnyPagePermission(WagtailTestUtils, TestCase):
    def test_superuser(self):
        user = self.create_superuser(
            username="superuser", email="admin@example.com", password="p"
        )
        self.assertTrue(user_has_any_page_permission(user))

    def test_inactive_superuser(self):
        user = self.create_superuser(
            username="superuser", email="admin@example.com", password="p"
        )
        user.is_active = False
        self.assertFalse(user_has_any_page_permission(user))

    def test_editor(self):
        user = self.create_user(username="editor", email="ed@example.com", password="p")
        editors = Group.objects.get(name="Editors")
        user.groups.add(editors)
        self.assertTrue(user_has_any_page_permission(user))

    def test_moderator(self):
        user = self.create_user(
            username="moderator", email="mod@example.com", password="p"
        )
        editors = Group.objects.get(name="Moderators")
        user.groups.add(editors)
        self.assertTrue(user_has_any_page_permission(user))

    def test_no_permissions(self):
        user = self.create_user(username="pleb", email="pleb@example.com", password="p")
        user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.assertFalse(user_has_any_page_permission(user))


class Test404(WagtailTestUtils, TestCase):
    def test_admin_404_template_used_append_slash_true(self):
        self.login()
        with self.settings(APPEND_SLASH=True):
            response = self.client.get("/admin/sdfgdsfgdsfgsdf", follow=True)

            # Check 404 error after CommonMiddleware redirect
            self.assertEqual(response.status_code, 404)
            self.assertTemplateUsed(response, "wagtailadmin/404.html")
            soup = self.get_soup(response.content)
            self.assertFalse(soup.select("script"))
            self.assertFalse(soup.select("[data-sprite]"))

    def test_not_logged_in_redirect(self):
        response = self.client.get("/admin/sdfgdsfgdsfgsdf/")

        # Check that the user was redirected to the login page and that next was set correctly
        self.assertRedirects(
            response, reverse("wagtailadmin_login") + "?next=/admin/sdfgdsfgdsfgsdf/"
        )


class TestAdminURLAppendSlash(WagtailTestUtils, TestCase):
    def setUp(self):
        # Find root page
        self.root_page = Page.objects.get(id=2)

    def test_return_correct_view_for_correct_url_without_ending_slash(self):
        self.login()
        with self.settings(APPEND_SLASH=True):
            # Remove trailing slash from URL
            response = self.client.get(
                reverse("wagtailadmin_explore_root")[:-1], follow=True
            )

            # Check that correct page is returned after CommonMiddleware redirect
            self.assertEqual(response.status_code, 200)
            self.assertTemplateUsed(
                response, "wagtailadmin/pages/explorable_index.html"
            )
            self.assertEqual(Page.objects.get(id=1), response.context["parent_page"])
            self.assertIn(self.root_page, response.context["pages"])


class TestRemoveStaleContentTypes(TestCase):
    def test_remove_stale_content_types_preserves_access_admin_permission(self):
        call_command("remove_stale_contenttypes", interactive=False)
        self.assertTrue(
            Permission.objects.filter(
                content_type__app_label="wagtailadmin", codename="access_admin"
            ).exists()
        )



# Generated by Django 2.2.10 on 2020-07-13 15:59

from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("wagtailcore", "0057_page_locale_fields_notnull"),
    ]

    operations = [
        migrations.CreateModel(
            name="TestModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "translation_key",
                    models.UUIDField(default=uuid.uuid4, editable=False),
                ),
                ("title", models.CharField(max_length=255)),
                (
                    "locale",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="wagtailcore.Locale",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("translation_key", "locale")},
            },
        ),
        migrations.CreateModel(
            name="TestPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wagtailcore.Page",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="InheritedTestModel",
            fields=[
                (
                    "testmodel_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="i18n.TestModel",
                    ),
                ),
            ],
            bases=("i18n.testmodel",),
        ),
        migrations.CreateModel(
            name="TestNonParentalChildObject",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "translation_key",
                    models.UUIDField(default=uuid.uuid4, editable=False),
                ),
                (
                    "sort_order",
                    models.IntegerField(blank=True, editable=False, null=True),
                ),
                ("field", models.TextField()),
                (
                    "locale",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="wagtailcore.Locale",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="test_nonparentalchildobjects",
                        to="i18n.TestPage",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("translation_key", "locale")},
            },
        ),
        migrations.CreateModel(
            name="TestChildObject",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "translation_key",
                    models.UUIDField(default=uuid.uuid4, editable=False),
                ),
                (
                    "sort_order",
                    models.IntegerField(blank=True, editable=False, null=True),
                ),
                ("field", models.TextField()),
                (
                    "locale",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="wagtailcore.Locale",
                    ),
                ),
                (
                    "page",
                    modelcluster.fields.ParentalKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="test_childobjects",
                        to="i18n.TestPage",
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
                "unique_together": {("translation_key", "locale")},
            },
        ),
    ]



from django.forms import Media

from wagtail import hooks
from wagtail.admin.auth import user_has_any_page_permission
from wagtail.admin.navigation import get_site_for_user
from wagtail.admin.ui.components import Component
from wagtail.models import Page, Site


class SummaryItem(Component):
    order = 100

    def __init__(self, request):
        self.request = request

    def is_shown(self):
        return True


class PagesSummaryItem(SummaryItem):
    order = 100
    template_name = "wagtailadmin/home/site_summary_pages.html"

    def get_context_data(self, parent_context):
        site_details = get_site_for_user(self.request.user)
        root_page = site_details["root_page"]
        site_name = site_details["site_name"]

        if root_page:
            page_count = Page.objects.descendant_of(root_page, inclusive=True).count()

            if root_page.is_root():
                # If the root page the user has access to is the Wagtail root,
                # subtract one from this count because the root is not a real page.
                page_count -= 1

                # If precisely one site exists, link to its homepage rather than the
                # tree root, to discourage people from trying to create pages as siblings
                # of the homepage (#1883)
                try:
                    root_page = Site.objects.get().root_page
                except (Site.DoesNotExist, Site.MultipleObjectsReturned):
                    pass
        else:
            page_count = 0

        return {
            "root_page": root_page,
            "total_pages": page_count,
            "site_name": site_name,
        }

    def is_shown(self):
        return user_has_any_page_permission(self.request.user)


class SiteSummaryPanel(Component):
    template_name = "wagtailadmin/home/site_summary.html"

    def __init__(self, request):
        self.request = request
        summary_items = []
        for fn in hooks.get_hooks("construct_homepage_summary_items"):
            fn(request, summary_items)
        self.summary_items = [s for s in summary_items if s.is_shown()]
        self.summary_items.sort(key=lambda p: p.order)

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        context["summary_items"] = self.summary_items
        return context

    @property
    def media(self):
        media = Media()
        for item in self.summary_items:
            media += item.media
        return media



import subprocess

from setuptools import setup
from setuptools.command.sdist import sdist as base_sdist


class sdist(base_sdist):
    def compile_assets(self):
        try:
            subprocess.check_call(["npm", "run", "build"])
        except (OSError, subprocess.CalledProcessError) as e:
            print("Error compiling assets: " + str(e))  # noqa: T201
            raise SystemExit(1)

    def run(self):
        self.compile_assets()
        super().run()


setup(
    name="wagtail",
    cmdclass={"sdist": sdist},
)



import logging
import uuid
from collections import defaultdict
from urllib.parse import urlparse

from django.core.exceptions import ImproperlyConfigured

from .base import BaseBackend

logger = logging.getLogger("wagtail.frontendcache")


__all__ = ["CloudfrontBackend"]


class CloudfrontBackend(BaseBackend):
    def __init__(self, params):
        import boto3

        super().__init__(params)

        self.client = boto3.client(
            "cloudfront",
            aws_access_key_id=params.get("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=params.get("AWS_SECRET_ACCESS_KEY"),
            aws_session_token=params.get("AWS_SESSION_TOKEN"),
        )

        try:
            self.cloudfront_distribution_id = params.pop("DISTRIBUTION_ID")
        except KeyError:
            raise ImproperlyConfigured(
                "The setting 'WAGTAILFRONTENDCACHE' requires the object 'DISTRIBUTION_ID'."
            )

    def purge_batch(self, urls):
        paths_by_distribution_id = defaultdict(set)

        for url in urls:
            url_parsed = urlparse(url)
            distribution_id = self.cloudfront_distribution_id

            if distribution_id:
                path = url_parsed.path or "/"
                if url_parsed.query:
                    path += "?" + url_parsed.query
                # Only append if the path is not already in the list
                if path not in paths_by_distribution_id[distribution_id]:
                    paths_by_distribution_id[distribution_id].add(path)

        for distribution_id, paths in paths_by_distribution_id.items():
            self._create_invalidation(distribution_id, list(paths))

    def purge(self, url):
        self.purge_batch([url])

    def _create_invalidation(self, distribution_id, paths):
        import botocore

        try:
            self.client.create_invalidation(
                DistributionId=distribution_id,
                InvalidationBatch={
                    "Paths": {"Quantity": len(paths), "Items": paths},
                    "CallerReference": str(uuid.uuid4()),
                },
            )
        except botocore.exceptions.ClientError as e:
            for path in paths:
                logger.error(
                    "Couldn't purge path '%s' from CloudFront (DistributionId=%s). ClientError: %s %s",
                    path,
                    distribution_id,
                    e.response["Error"]["Code"],
                    e.response["Error"]["Message"],
                )



# Generated by Django 4.2.dev20221212151407 on 2022-12-12 15:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("wagtailcore", "0081_populate_workflowstate_content_type"),
    ]

    operations = [
        migrations.AlterField(
            model_name="workflowstate",
            name="base_content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="workflowstate",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="contenttypes.contenttype",
            ),
        ),
    ]



from datetime import timedelta
from io import StringIO
from unittest import mock

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core import management
from django.core.cache import cache
from django.db import models
from django.test import TestCase, override_settings
from django.utils import timezone

from wagtail.embeds.models import Embed
from wagtail.models import (
    Collection,
    Page,
    PageLogEntry,
    Revision,
    Task,
    Workflow,
    WorkflowTask,
)
from wagtail.signals import page_published, page_unpublished, published, unpublished
from wagtail.test.testapp.models import (
    DraftStateModel,
    EventPage,
    FullFeaturedSnippet,
    PurgeRevisionsProtectedTestModel,
    SecretPage,
    SimplePage,
)
from wagtail.test.utils import WagtailTestUtils


class TestFixTreeCommand(TestCase):
    fixtures = ["test.json"]

    def badly_delete_page(self, page):
        # Deletes a page the wrong way.
        # This will not update numchild and may leave orphans
        models.Model.delete(page)

    def run_command(self, **options):
        options.setdefault("interactive", False)

        output = StringIO()
        management.call_command("fixtree", stdout=output, **options)
        output.seek(0)

        return output

    def test_fixes_numchild(self):
        # Get homepage and save old value
        homepage = Page.objects.get(url_path="/home/")
        old_numchild = homepage.numchild

        # Break it
        homepage.numchild = 12345
        homepage.save()

        # Check that its broken
        self.assertEqual(Page.objects.get(url_path="/home/").numchild, 12345)

        # Call command
        self.run_command()

        # Check if its fixed
        self.assertEqual(Page.objects.get(url_path="/home/").numchild, old_numchild)

    def test_fixes_depth(self):
        # Get homepage and save old value
        homepage = Page.objects.get(url_path="/home/")
        old_depth = homepage.depth

        # Break it
        homepage.depth = 12345
        homepage.save()

        # also break the root collection's depth
        root_collection = Collection.get_first_root_node()
        root_collection.depth = 42
        root_collection.save()

        # Check that its broken
        self.assertEqual(Page.objects.get(url_path="/home/").depth, 12345)
        self.assertEqual(Collection.objects.get(id=root_collection.id).depth, 42)

        # Call command
        self.run_command()

        # Check if its fixed
        self.assertEqual(Page.objects.get(url_path="/home/").depth, old_depth)
        self.assertEqual(Collection.objects.get(id=root_collection.id).depth, 1)

    def test_detects_orphans(self):
        events_index = Page.objects.get(url_path="/home/events/")
        christmas_page = EventPage.objects.get(url_path="/home/events/christmas/")

        # Delete the events index badly
        self.badly_delete_page(events_index)

        # Check that christmas_page is still in the tree
        self.assertTrue(Page.objects.filter(id=christmas_page.id).exists())

        # Call command
        output = self.run_command()

        # Check that the issues were detected
        output_string = output.read()
        self.assertIn("Incorrect numchild value found for pages: [2]", output_string)
        # Note that page ID 15 was also deleted, but is not picked up here, as
        # it is a child of 14.
        self.assertIn("Orphaned pages found: [4, 5, 6, 9, 13, 15]", output_string)

        # Check that christmas_page is still in the tree
        self.assertTrue(Page.objects.filter(id=christmas_page.id).exists())

    def test_deletes_orphans(self):
        events_index = Page.objects.get(url_path="/home/events/")
        christmas_page = EventPage.objects.get(url_path="/home/events/christmas/")

        # Delete the events index badly
        self.badly_delete_page(events_index)

        # Check that christmas_page is still in the tree
        self.assertTrue(Page.objects.filter(id=christmas_page.id).exists())

        # Call command
        # delete_orphans simulates a user pressing "y" at the prompt
        output = self.run_command(delete_orphans=True)

        # Check that the issues were detected
        output_string = output.read()
        self.assertIn("Incorrect numchild value found for pages: [2]", output_string)
        self.assertIn("7 orphaned pages deleted.", output_string)

        # Check that christmas_page has been deleted
        self.assertFalse(Page.objects.filter(id=christmas_page.id).exists())

    def test_remove_path_holes(self):
        events_index = Page.objects.get(url_path="/home/events/")
        # Delete the event page in path position 0001
        Page.objects.get(path=events_index.path + "0001").delete()

        self.run_command(full=True)
        # the gap at position 0001 should have been closed
        events_index = Page.objects.get(url_path="/home/events/")
        self.assertTrue(Page.objects.filter(path=events_index.path + "0001").exists())


class TestMovePagesCommand(TestCase):
    fixtures = ["test.json"]

    def run_command(self, from_, to):
        management.call_command("move_pages", str(from_), str(to), stdout=StringIO())

    def test_move_pages(self):
        # Get pages
        events_index = Page.objects.get(url_path="/home/events/")
        about_us = Page.objects.get(url_path="/home/about-us/")
        page_ids = events_index.get_children().values_list("id", flat=True)

        # Move all events into "about us"
        self.run_command(events_index.id, about_us.id)

        # Check that all pages moved
        for page_id in page_ids:
            self.assertEqual(Page.objects.get(id=page_id).get_parent(), about_us)


class TestSetUrlPathsCommand(TestCase):
    fixtures = ["test.json"]

    def run_command(self):
        management.call_command("set_url_paths", stdout=StringIO())

    def test_set_url_paths(self):
        self.run_command()


@override_settings(
    CACHES={
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        },
    }
)
class TestPublishScheduledPagesCommand(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        cache.clear()

        # Find root page
        self.root_page = Page.objects.get(id=2)

    def test_go_live_page_will_be_published(self):
        # Connect a mock signal handler to page_published signal
        signal_fired = [False]
        signal_page = [None]

        def page_published_handler(sender, instance, **kwargs):
            signal_fired[0] = True
            signal_page[0] = instance

        page_published.connect(page_published_handler)

        try:
            page = SimplePage(
                title="Hello world!",
                slug="hello-world",
                content="hello",
                live=False,
                has_unpublished_changes=True,
                go_live_at=timezone.now() - timedelta(days=1),
            )
            self.root_page.add_child(instance=page)

            page.save_revision(approved_go_live_at=timezone.now() - timedelta(days=1))

            p = Page.objects.get(slug="hello-world")
            self.assertFalse(p.live)
            self.assertTrue(
                Revision.page_revisions.filter(object_id=p.id)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )
            with self.assertNumQueries(49):
                with self.captureOnCommitCallbacks(execute=True):
                    management.call_command("publish_scheduled_pages")

            p = Page.objects.get(slug="hello-world")
            self.assertTrue(p.live)
            self.assertTrue(p.first_published_at)
            self.assertFalse(p.has_unpublished_changes)
            self.assertFalse(
                Revision.page_revisions.filter(object_id=p.id)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            # Check that the page_published signal was fired
            self.assertTrue(signal_fired[0])
            self.assertEqual(signal_page[0], page)
            self.assertEqual(signal_page[0], signal_page[0].specific)
        finally:
            page_published.disconnect(page_published_handler)

    def test_go_live_page_created_by_editor_will_be_published(self):
        # Connect a mock signal handler to page_published signal
        signal_fired = [False]
        signal_page = [None]

        editor = self.create_user("ed")
        editor.groups.add(Group.objects.get(name="Site-wide editors"))

        def page_published_handler(sender, instance, **kwargs):
            signal_fired[0] = True
            signal_page[0] = instance

        page_published.connect(page_published_handler)

        try:
            page = SimplePage(
                title="Hello world!",
                slug="hello-world",
                content="hello",
                live=False,
                has_unpublished_changes=True,
                go_live_at=timezone.now() - timedelta(days=1),
            )
            self.root_page.add_child(instance=page)

            page.save_revision(
                user=editor, approved_go_live_at=timezone.now() - timedelta(days=1)
            )

            p = Page.objects.get(slug="hello-world")
            self.assertFalse(p.live)
            self.assertTrue(
                Revision.page_revisions.filter(object_id=p.id)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            with self.assertNumQueries(49):
                with self.captureOnCommitCallbacks(execute=True):
                    management.call_command("publish_scheduled_pages")

            p = Page.objects.get(slug="hello-world")
            self.assertTrue(p.live)
            self.assertTrue(p.first_published_at)
            self.assertFalse(p.has_unpublished_changes)
            self.assertFalse(
                Revision.page_revisions.filter(object_id=p.id)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            # Check that the page_published signal was fired
            self.assertTrue(signal_fired[0])
            self.assertEqual(signal_page[0], page)
            self.assertEqual(signal_page[0], signal_page[0].specific)
        finally:
            page_published.disconnect(page_published_handler)

    def test_go_live_when_newer_revision_exists(self):
        page = SimplePage(
            title="Hello world!",
            slug="hello-world",
            content="hello",
            live=False,
            has_unpublished_changes=True,
            go_live_at=timezone.now() - timedelta(days=1),
        )
        self.root_page.add_child(instance=page)

        page.save_revision(approved_go_live_at=timezone.now() - timedelta(days=1))

        page.title = "Goodbye world!"
        page.save_revision()

        with self.assertNumQueries(49):
            with self.captureOnCommitCallbacks(execute=True):
                management.call_command("publish_scheduled_pages")

        p = Page.objects.get(slug="hello-world")
        self.assertTrue(p.live)
        self.assertTrue(p.has_unpublished_changes)
        self.assertEqual(p.title, "Hello world!")

    def test_future_go_live_page_will_not_be_published(self):
        page = SimplePage(
            title="Hello world!",
            slug="hello-world",
            content="hello",
            live=False,
            go_live_at=timezone.now() + timedelta(days=1),
        )
        self.root_page.add_child(instance=page)

        page.save_revision(approved_go_live_at=timezone.now() - timedelta(days=1))

        p = Page.objects.get(slug="hello-world")
        self.assertFalse(p.live)
        self.assertTrue(
            Revision.page_revisions.filter(object_id=p.id)
            .exclude(approved_go_live_at__isnull=True)
            .exists()
        )

        with self.assertNumQueries(42):
            with self.captureOnCommitCallbacks(execute=True):
                management.call_command("publish_scheduled_pages")

        p = Page.objects.get(slug="hello-world")
        self.assertFalse(p.live)
        self.assertTrue(
            Revision.page_revisions.filter(object_id=p.id)
            .exclude(approved_go_live_at__isnull=True)
            .exists()
        )

    def test_expired_page_will_be_unpublished(self):
        # Connect a mock signal handler to page_unpublished signal
        signal_fired = [False]
        signal_page = [None]

        def page_unpublished_handler(sender, instance, **kwargs):
            signal_fired[0] = True
            signal_page[0] = instance

        page_unpublished.connect(page_unpublished_handler)

        try:
            page = SimplePage(
                title="Hello world!",
                slug="hello-world",
                content="hello",
                live=True,
                has_unpublished_changes=False,
                expire_at=timezone.now() - timedelta(days=1),
            )
            self.root_page.add_child(instance=page)

            p = Page.objects.get(slug="hello-world")
            self.assertTrue(p.live)

            with self.assertNumQueries(29):
                with self.captureOnCommitCallbacks(execute=True):
                    management.call_command("publish_scheduled_pages")

            p = Page.objects.get(slug="hello-world")
            self.assertFalse(p.live)
            self.assertTrue(p.has_unpublished_changes)
            self.assertTrue(p.expired)

            # Check that the page_published signal was fired
            self.assertTrue(signal_fired[0])
            self.assertEqual(signal_page[0], page)
            self.assertEqual(signal_page[0], signal_page[0].specific)
        finally:
            page_unpublished.disconnect(page_unpublished_handler)

    def test_future_expired_page_will_not_be_unpublished(self):
        page = SimplePage(
            title="Hello world!",
            slug="hello-world",
            content="hello",
            live=True,
            expire_at=timezone.now() + timedelta(days=1),
        )
        self.root_page.add_child(instance=page)

        p = Page.objects.get(slug="hello-world")
        self.assertTrue(p.live)

        with self.assertNumQueries(6):
            with self.captureOnCommitCallbacks(execute=True):
                management.call_command("publish_scheduled_pages")

        p = Page.objects.get(slug="hello-world")
        self.assertTrue(p.live)
        self.assertFalse(p.expired)


@override_settings(
    CACHES={
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        },
    }
)
class TestPublishScheduledCommand(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        cache.clear()
        self.snippet = DraftStateModel.objects.create(text="Hello world!", live=False)

    def test_go_live_will_be_published(self):
        # Connect a mock signal handler to published signal
        signal_fired = [False]
        signal_obj = [None]

        def published_handler(sender, instance, **kwargs):
            signal_fired[0] = True
            signal_obj[0] = instance

        published.connect(published_handler)

        try:
            go_live_at = timezone.now() - timedelta(days=1)
            self.snippet.has_unpublished_changes = True
            self.snippet.go_live_at = go_live_at

            self.snippet.save_revision(approved_go_live_at=go_live_at)

            self.snippet.refresh_from_db()
            self.assertFalse(self.snippet.live)
            self.assertTrue(
                Revision.objects.for_instance(self.snippet)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            with self.assertNumQueries(15):
                with self.captureOnCommitCallbacks(execute=True):
                    management.call_command("publish_scheduled")

            self.snippet.refresh_from_db()
            self.assertTrue(self.snippet.live)
            self.assertTrue(self.snippet.first_published_at)
            self.assertFalse(self.snippet.has_unpublished_changes)
            self.assertFalse(
                Revision.objects.for_instance(self.snippet)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            # Check that the published signal was fired
            self.assertTrue(signal_fired[0])
            self.assertEqual(signal_obj[0], self.snippet)
        finally:
            published.disconnect(published_handler)

    def test_go_live_created_by_editor_will_be_published(self):
        # Connect a mock signal handler to published signal
        signal_fired = [False]
        signal_obj = [None]

        editor = self.create_user("ed")
        editor.groups.add(Group.objects.get(name="Site-wide editors"))

        def published_handler(sender, instance, **kwargs):
            signal_fired[0] = True
            signal_obj[0] = instance

        published.connect(published_handler)

        try:
            go_live_at = timezone.now() - timedelta(days=1)
            self.snippet.has_unpublished_changes = True
            self.snippet.go_live_at = go_live_at

            self.snippet.save_revision(user=editor, approved_go_live_at=go_live_at)

            self.snippet.refresh_from_db()
            self.assertFalse(self.snippet.live)
            self.assertTrue(
                Revision.objects.for_instance(self.snippet)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            with self.assertNumQueries(15):
                with self.captureOnCommitCallbacks(execute=True):
                    management.call_command("publish_scheduled")

            self.snippet.refresh_from_db()
            self.assertTrue(self.snippet.live)
            self.assertTrue(self.snippet.first_published_at)
            self.assertFalse(self.snippet.has_unpublished_changes)
            self.assertFalse(
                Revision.objects.for_instance(self.snippet)
                .exclude(approved_go_live_at__isnull=True)
                .exists()
            )

            # Check that the published signal was fired
            self.assertTrue(signal_fired[0])
            self.assertEqual(signal_obj[0], self.snippet)
        finally:
            published.disconnect(published_handler)

    def test_go_live_when_newer_revision_exists(self):
        go_live_at = timezone.now() - timedelta(days=1)
        self.snippet.has_unpublished_changes = True
        self.snippet.go_live_at = go_live_at

        self.snippet.save_revision(approved_go_live_at=go_live_at)

        self.snippet.text = "Goodbye world!"
        self.snippet.save_revision()

        with self.assertNumQueries(15):
            with self.captureOnCommitCallbacks(execute=True):
                management.call_command("publish_scheduled")

        self.snippet.refresh_from_db()
        self.assertTrue(self.snippet.live)
        self.assertTrue(self.snippet.has_unpublished_changes)
        self.assertEqual(self.snippet.text, "Hello world!")

    def test_future_go_live_will_not_be_published(self):
        self.snippet.has_unpublished_changes = True
        self.snippet.go_live_at = timezone.now() + timedelta(days=1)

        self.snippet.save_revision(
            approved_go_live_at=timezone.now() - timedelta(days=1)
        )

        self.snippet.refresh_from_db()
        self.assertFalse(self.snippet.live)
        self.assertTrue(
            Revision.objects.for_instance(self.snippet)
            .exclude(approved_go_live_at__isnull=True)
            .exists()
        )

        with self.assertNumQueries(14):
            with self.captureOnCommitCallbacks(execute=True):
                management.call_command("publish_scheduled")

        self.assertFalse(self.snippet.live)
        self.assertTrue(
            Revision.objects.for_instance(self.snippet)
            .exclude(approved_go_live_at__isnull=True)
            .exists()
        )

    def test_expired_will_be_unpublished(self):
        # Connect a mock signal handler to unpublished signal
        signal_fired = [False]
        signal_obj = [None]

        def unpublished_handler(sender, instance, **kwargs):
            signal_fired[0] = True
            signal_obj[0] = instance

        unpublished.connect(unpublished_handler)

        try:
            self.snippet.expire_at = timezone.now() - timedelta(days=1)
            self.snippet.save_revision().publish()

            self.snippet.refresh_from_db()
            self.assertTrue(self.snippet.live)

            with self.assertNumQueries(10):
                with self.captureOnCommitCallbacks(execute=True):
                    management.call_command("publish_scheduled")

            self.snippet.refresh_from_db()
            self.assertFalse(self.snippet.live)
            self.assertTrue(self.snippet.has_unpublished_changes)
            self.assertTrue(self.snippet.expired)

            # Check that the unpublished signal was fired
            self.assertTrue(signal_fired[0])
            self.assertEqual(signal_obj[0], self.snippet)
        finally:
            unpublished.disconnect(unpublished_handler)

    def test_future_expired_will_not_be_unpublished(self):
        self.snippet.expire_at = timezone.now() + timedelta(days=1)
        self.snippet.save_revision().publish()

        self.snippet.refresh_from_db()
        self.assertTrue(self.snippet.live)

        with self.assertNumQueries(6):
            with self.captureOnCommitCallbacks(execute=True):
                management.call_command("publish_scheduled")

        self.snippet.refresh_from_db()
        self.assertTrue(self.snippet.live)
        self.assertFalse(self.snippet.expired)


class TestPurgeRevisionsCommandForPages(TestCase):
    base_options = {}

    def setUp(self):
        self.object = self.get_object()

    def get_object(self):
        # Find root page
        self.root_page = Page.objects.get(id=2)
        self.page = SimplePage(
            title="Hello world!",
            slug="hello-world",
            content="hello",
            live=False,
        )
        self.root_page.add_child(instance=self.page)
        self.page.refresh_from_db()
        return self.page

    def assertRevisionNotExists(self, revision):
        self.assertFalse(Revision.objects.filter(id=revision.id).exists())

    def assertRevisionExists(self, revision):
        self.assertTrue(Revision.objects.filter(id=revision.id).exists())

    def run_command(self, **options):
        return management.call_command(
            "purge_revisions", **{**self.base_options, **options}, stdout=StringIO()
        )

    def test_latest_revision_not_purged(self):
        revision_1 = self.object.save_revision()
        revision_2 = self.object.save_revision()

        self.run_command()

        # revision 1 should be deleted, revision 2 should not be
        self.assertRevisionNotExists(revision_1)
        self.assertRevisionExists(revision_2)

    def test_revisions_in_moderation_or_workflow_not_purged(self):
        workflow = Workflow.objects.create(name="test_workflow")
        task_1 = Task.objects.create(name="test_task_1")
        user = get_user_model().objects.first()
        WorkflowTask.objects.create(workflow=workflow, task=task_1, sort_order=1)

        revision = self.object.save_revision()
        workflow.start(self.object, user)

        # Save a new revision to ensure that the revision in the workflow
        # is not the latest one
        self.object.save_revision()

        self.run_command()

        # even though they're no longer the latest revisions, the old revisions
        # should stay as they are attached to an in progress workflow
        self.assertRevisionExists(revision)

        # If workflow is disabled at some point after that, the revision should
        # be deleted
        with override_settings(WAGTAIL_WORKFLOW_ENABLED=False):
            self.run_command()
            self.assertRevisionNotExists(revision)

    def test_revisions_with_approve_go_live_not_purged(self):
        revision = self.object.save_revision(
            approved_go_live_at=timezone.now() + timedelta(days=1)
        )

        # Save a new revision to ensure that the approved revision
        # is not the latest one
        self.object.save_revision()

        self.run_command()

        self.assertRevisionExists(revision)

    def test_purge_revisions_with_date_cutoff(self):
        old_revision = self.object.save_revision()

        self.object.save_revision()

        self.run_command(days=30)

        # revision should not be deleted, as it is younger than 30 days
        self.assertRevisionExists(old_revision)

        old_revision.created_at = timezone.now() - timedelta(days=31)
        old_revision.save()

        self.run_command(days=30)

        # revision is now older than 30 days, so should be deleted
        self.assertRevisionNotExists(old_revision)

    def test_purge_revisions_protected_error(self):
        revision_old = self.object.save_revision()
        PurgeRevisionsProtectedTestModel.objects.create(revision=revision_old)
        revision_purged = self.object.save_revision()
        self.object.save_revision()

        self.run_command()
        # revision should not be deleted, as it is protected
        self.assertRevisionExists(revision_old)
        # Any other revisions are deleted
        self.assertRevisionNotExists(revision_purged)


class TestPurgeRevisionsCommandForSnippets(TestPurgeRevisionsCommandForPages):
    def get_object(self):
        return FullFeaturedSnippet.objects.create(text="Hello world!")


class TestPurgeRevisionsCommandForPagesWithPagesOnly(TestPurgeRevisionsCommandForPages):
    base_options = {"pages": True}


class TestPurgeRevisionsCommandForPagesWithNonPagesOnly(
    TestPurgeRevisionsCommandForPages
):
    base_options = {"non_pages": True}

    def assertRevisionNotExists(self, revision):
        # Page revisions won't be purged if only non_pages is specified
        return self.assertRevisionExists(revision)


class TestPurgeRevisionsCommandForSnippetsWithNonPagesOnly(
    TestPurgeRevisionsCommandForSnippets
):
    base_options = {"non_pages": True}


class TestPurgeRevisionsCommandForSnippetsWithPagesOnly(
    TestPurgeRevisionsCommandForSnippets
):
    base_options = {"pages": True}

    def assertRevisionNotExists(self, revision):
        # Snippet revisions won't be purged if only pages is specified
        return self.assertRevisionExists(revision)


class TestPurgeEmbedsCommand(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        # create dummy Embed objects
        for i in range(5):
            embed = Embed(
                hash=f"{i}",
                url="https://www.youtube.com/watch?v=Js8dIRxwSRY",
                max_width=None,
                type="video",
                html="test html",
                title="test title",
                author_name="test author name",
                provider_name="test provider name",
                thumbnail_url="http://test/thumbnail.url",
                width=1000,
                height=1000,
            )
            embed.save()

    def test_purge_embeds(self):
        """
        fetch all dummy embeds and confirm they are deleted when the management command runs

        """

        self.assertEqual(Embed.objects.count(), 5)

        management.call_command("purge_embeds", stdout=StringIO())

        self.assertEqual(Embed.objects.count(), 0)


class TestCreateLogEntriesFromRevisionsCommand(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        self.page = SimplePage(
            title="Hello world!",
            slug="hello-world",
            content="hello",
            live=False,
            expire_at=timezone.now() - timedelta(days=1),
        )

        Page.objects.get(id=2).add_child(instance=self.page)

        # Create empty revisions, which should not be converted to log entries
        for i in range(3):
            self.page.save_revision()

        # Add another revision with a content change
        self.page.title = "Hello world!!"
        revision = self.page.save_revision()
        revision.publish()

        # Do the same with a SecretPage (to check that the version comparison code doesn't
        # trip up on permission-dependent edit handlers)
        self.secret_page = SecretPage(
            title="The moon",
            slug="the-moon",
            boring_data="the moon",
            secret_data="is made of cheese",
            live=False,
        )

        Page.objects.get(id=2).add_child(instance=self.secret_page)

        # Create empty revisions, which should not be converted to log entries
        for i in range(3):
            self.secret_page.save_revision()

        # Add another revision with a content change
        self.secret_page.secret_data = "is flat"
        revision = self.secret_page.save_revision()
        revision.publish()

        # clean up log entries
        PageLogEntry.objects.all().delete()

    def test_log_entries_created_from_revisions(self):
        management.call_command("create_log_entries_from_revisions")

        # Should not create entries for empty revisions.
        self.assertListEqual(
            list(PageLogEntry.objects.values_list("page_id", "action")),
            # Default PageLogEntry sort order is from newest event to oldest.
            # We reverse here to make it easier to understand what is being
            # tested. The events here should correspond with setUp above.
            list(
                reversed(
                    [
                        # The SimplePage was created in draft mode, with an initial revision.
                        (self.page.pk, "wagtail.create"),
                        (self.page.pk, "wagtail.edit"),
                        # The SimplePage was edited as a new draft, then published.
                        (self.page.pk, "wagtail.edit"),
                        (self.page.pk, "wagtail.publish"),
                        # The SecretPage was created in draft mode, with an initial revision.
                        (self.secret_page.pk, "wagtail.create"),
                        (self.secret_page.pk, "wagtail.edit"),
                        # The SecretPage was edited as a new draft, then published.
                        (self.secret_page.pk, "wagtail.edit"),
                        (self.secret_page.pk, "wagtail.publish"),
                    ]
                )
            ),
        )

    def test_command_doesnt_crash_for_revisions_without_page_model(self):
        with mock.patch(
            "wagtail.models.Page.specific_class",
            return_value=None,
            new_callable=mock.PropertyMock,
        ):
            management.call_command("create_log_entries_from_revisions")
            self.assertEqual(PageLogEntry.objects.count(), 0)



# Generated by Django 4.2.dev20221212120043 on 2022-12-12 13:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0078_referenceindex"),
    ]

    operations = [
        migrations.RenameField(
            model_name="taskstate",
            old_name="page_revision",
            new_name="revision",
        ),
    ]



from django import forms
from django.core.validators import URLValidator
from django.utils.translation import gettext_lazy as _


class EmbedForm(forms.Form):
    url = forms.CharField(
        label=_("URL"), validators=[URLValidator(message=_("Please enter a valid URL"))]
    )



from collections import OrderedDict

import django.forms
from django.conf import settings
from django.utils.html import conditional_escape
from django.utils.translation import gettext_lazy as _

from wagtail.admin.forms import WagtailAdminPageForm
from wagtail.compat import URLField


def _split_by_newline_or_comma(value):
    """
    Split the given string either by new lines, or if no new line is present then
    by comma.
    """
    if len(lines := value.strip().splitlines()) > 1:
        return [line.strip().rstrip(",").strip() for line in lines]
    else:
        return list(map(str.strip, value.split(",")))


class BaseForm(django.forms.Form):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("label_suffix", "")

        self.user = kwargs.pop("user", None)
        self.page = kwargs.pop("page", None)

        super().__init__(*args, **kwargs)


class FormBuilder:
    def __init__(self, fields):
        self.fields = fields

    def create_singleline_field(self, field, options):
        # TODO: This is a default value - it may need to be changed
        options["max_length"] = 255
        return django.forms.CharField(**options)

    def create_multiline_field(self, field, options):
        options.setdefault("widget", django.forms.Textarea)
        return django.forms.CharField(**options)

    def create_date_field(self, field, options):
        return django.forms.DateField(**options)

    def create_datetime_field(self, field, options):
        return django.forms.DateTimeField(**options)

    def create_email_field(self, field, options):
        return django.forms.EmailField(**options)

    def create_url_field(self, field, options):
        return URLField(**options)

    def create_number_field(self, field, options):
        return django.forms.DecimalField(**options)

    def create_dropdown_field(self, field, options):
        options["choices"] = self.get_formatted_field_choices(field)
        return django.forms.ChoiceField(**options)

    def create_multiselect_field(self, field, options):
        options["choices"] = self.get_formatted_field_choices(field)
        return django.forms.MultipleChoiceField(**options)

    def create_radio_field(self, field, options):
        options["choices"] = self.get_formatted_field_choices(field)
        options.setdefault("widget", django.forms.RadioSelect)
        return django.forms.ChoiceField(**options)

    def create_checkboxes_field(self, field, options):
        options["choices"] = self.get_formatted_field_choices(field)
        options["initial"] = self.get_formatted_field_initial(field)
        options.setdefault("widget", django.forms.CheckboxSelectMultiple)
        return django.forms.MultipleChoiceField(**options)

    def create_checkbox_field(self, field, options):
        return django.forms.BooleanField(**options)

    def create_hidden_field(self, field, options):
        options.setdefault("widget", django.forms.HiddenInput)
        return django.forms.CharField(**options)

    def get_create_field_function(self, type):
        """
        Takes string of field type and returns a Django Form Field Instance.
        Assumes form field creation functions are in the format:
        'create_fieldtype_field'
        """
        create_field_function = getattr(self, "create_%s_field" % type, None)
        if create_field_function:
            return create_field_function
        else:
            import inspect

            method_list = [
                f[0]
                for f in inspect.getmembers(self.__class__, inspect.isfunction)
                if f[0].startswith("create_") and f[0].endswith("_field")
            ]
            raise AttributeError(
                "Could not find function matching format \
                create_<fieldname>_field for type: "
                + type,
                "Must be one of: " + ", ".join(method_list),
            )

    def get_formatted_field_choices(self, field):
        """
        Returns a list of choices [(string, string),] for the field.
        Split the provided choices into a list, separated by new lines.
        If no new lines in the provided choices, split by commas.
        """
        return [(x, x) for x in _split_by_newline_or_comma(field.choices)]

    def get_formatted_field_initial(self, field):
        """
        Returns a list of initial values [string,] for the field.
        Split the supplied default values into a list, separated by new lines.
        If no new lines in the provided default values, split by commas.
        """
        return _split_by_newline_or_comma(field.default_value)

    @property
    def formfields(self):
        formfields = OrderedDict()

        for field in self.fields:
            options = self.get_field_options(field)
            create_field = self.get_create_field_function(field.field_type)

            # If the field hasn't been saved to the database yet (e.g. we are previewing
            # a FormPage with unsaved changes) it won't have a clean_name as this is
            # set in FormField.save.
            clean_name = field.clean_name or field.get_field_clean_name()
            formfields[clean_name] = create_field(field, options)

        return formfields

    def get_field_options(self, field):
        options = {"label": field.label}
        if getattr(settings, "WAGTAILFORMS_HELP_TEXT_ALLOW_HTML", False):
            options["help_text"] = field.help_text
        else:
            options["help_text"] = conditional_escape(field.help_text)
        options["required"] = field.required
        options["initial"] = field.default_value
        return options

    def get_form_class(self):
        return type("WagtailForm", (BaseForm,), self.formfields)


class SelectDateForm(django.forms.Form):
    date_from = django.forms.DateTimeField(
        required=False,
        widget=django.forms.DateInput(attrs={"placeholder": _("Date from")}),
    )
    date_to = django.forms.DateTimeField(
        required=False,
        widget=django.forms.DateInput(attrs={"placeholder": _("Date to")}),
    )


class WagtailAdminFormPageForm(WagtailAdminPageForm):
    def clean(self):
        """
        Dynamically detect all related AbstractFormField subclasses to ensure
        validation is applied regardless of the related_name or if there are multiple
        AbstractFormField subclasses related to this page.
        """

        from .models import AbstractFormField

        cleaned_data = super().clean()

        form_fields_related_names = [
            related_object.related_name
            for related_object in self.instance._meta.related_objects
            if issubclass(related_object.related_model, AbstractFormField)
        ]

        for related_name in form_fields_related_names:
            if related_name not in self.formsets:
                continue

            forms = self.formsets[related_name].forms
            for form in forms:
                form.is_valid()

            seen_names = set()
            duplicate_names = set()

            # Find duplicates
            for form in forms:
                if form.cleaned_data.get("DELETE", False):
                    continue

                # Use existing clean_name or generate for new fields,
                # raise an error if there are duplicate resolved clean names.
                # Note: `clean_name` is set in `FormField.save`.
                clean_name = (
                    form.instance.clean_name or form.instance.get_field_clean_name()
                )

                if clean_name in seen_names:
                    duplicate_names.add(clean_name)
                else:
                    seen_names.add(clean_name)

            # Add validation errors to forms with duplicate clean names
            if duplicate_names:
                for form in forms:
                    if form.cleaned_data.get("DELETE", False):
                        continue

                    clean_name = (
                        form.instance.clean_name or form.instance.get_field_clean_name()
                    )

                    if clean_name in duplicate_names:
                        if form.instance.clean_name:
                            form.add_error(
                                "label",
                                django.forms.ValidationError(
                                    _(
                                        "The name '%(duplicate_clean_name)s' is "
                                        "already assigned to this field and cannot be "
                                        "changed. Please enter a different label for "
                                        "the conflicting field."
                                    )
                                    % {"duplicate_clean_name": clean_name}
                                ),
                            )
                        else:
                            form.add_error(
                                "label",
                                django.forms.ValidationError(
                                    _(
                                        "Field name '%(duplicate_clean_name)s' "
                                        "conflicts with another field. Please "
                                        "change the label."
                                    )
                                    % {"duplicate_clean_name": clean_name}
                                ),
                            )

        return cleaned_data



from django.contrib.auth import get_permission_codename
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.forms.models import modelform_factory
from django.urls import path
from django.utils.functional import cached_property
from django.utils.text import capfirst

from wagtail import hooks
from wagtail.admin.admin_url_finder import (
    ModelAdminURLFinder,
    register_admin_url_finder,
)
from wagtail.admin.panels.group import ObjectList
from wagtail.admin.views import generic
from wagtail.admin.views.generic import history, usage
from wagtail.models import ReferenceIndex
from wagtail.permissions import ModelPermissionPolicy

from .base import ViewSet, ViewSetGroup


class ModelViewSet(ViewSet):
    """
    A viewset to allow listing, creating, editing and deleting model instances.

    All attributes and methods from :class:`~wagtail.admin.viewsets.base.ViewSet`
    are available.

    For more information on how to use this class, see :ref:`generic_views`.
    """

    #: Register the model to the reference index to track its usage.
    #: For more details, see :ref:`managing_the_reference_index`.
    add_to_reference_index = True

    #: The view class to use for the index view; must be a subclass of ``wagtail.admin.views.generic.IndexView``.
    index_view_class = generic.IndexView

    #: The view class to use for the create view; must be a subclass of ``wagtail.admin.views.generic.CreateView``.
    add_view_class = generic.CreateView

    #: The view class to use for the edit view; must be a subclass of ``wagtail.admin.views.generic.EditView``.
    edit_view_class = generic.EditView

    #: The view class to use for the delete view; must be a subclass of ``wagtail.admin.views.generic.DeleteView``.
    delete_view_class = generic.DeleteView

    #: The view class to use for the history view; must be a subclass of ``wagtail.admin.views.generic.history.HistoryView``.
    history_view_class = history.HistoryView

    #: The view class to use for the usage view; must be a subclass of ``wagtail.admin.views.generic.usage.UsageView``.
    usage_view_class = usage.UsageView

    #: The view class to use for the copy view; must be a subclass of ``wagtail.admin.views.generic.CopyView``.
    copy_view_class = generic.CopyView

    #: The view class to use for the inspect view; must be a subclass of ``wagtail.admin.views.generic.InspectView``.
    inspect_view_class = generic.InspectView

    #: The prefix of template names to look for when rendering the admin views.
    template_prefix = ""

    #: The number of items to display per page in the index view. Defaults to 20.
    list_per_page = 20

    #: The default ordering to use for the index view.
    #: Can be a string or a list/tuple in the same format as Django's
    #: :attr:`~django.db.models.Options.ordering`.
    ordering = None

    #: Whether to enable the inspect view. Defaults to ``False``.
    inspect_view_enabled = False

    #: The model fields or attributes to display in the inspect view.
    #:
    #: If the field has a corresponding :meth:`~django.db.models.Model.get_FOO_display`
    #: method on the model, the method's return value will be used instead.
    #:
    #: If you have ``wagtail.images`` installed, and the field's value is an instance of
    #: ``wagtail.images.models.AbstractImage``, a thumbnail of that image will be rendered.
    #:
    #: If you have ``wagtail.documents`` installed, and the field's value is an instance of
    #: ``wagtail.docs.models.AbstractDocument``, a link to that document will be rendered,
    #: along with the document title, file extension and size.
    inspect_view_fields = []

    #: The fields to exclude from the inspect view.
    inspect_view_fields_exclude = []

    #: Whether to enable the copy view. Defaults to ``True``.
    copy_view_enabled = True

    def __init__(self, name=None, **kwargs):
        super().__init__(name=name, **kwargs)
        if not self.model:
            raise ImproperlyConfigured(
                "ModelViewSet %r must define a `model` attribute or pass a `model` argument"
                % self
            )

        self.model_opts = self.model._meta
        self.app_label = self.model_opts.app_label
        self.model_name = self.model_opts.model_name

    @property
    def permission_policy(self):
        return ModelPermissionPolicy(self.model)

    @cached_property
    def name(self):
        """
        Viewset name, to use as the URL prefix and namespace.
        Defaults to the :attr:`~django.db.models.Options.model_name`.
        """
        return self.model_name

    def get_common_view_kwargs(self, **kwargs):
        view_kwargs = super().get_common_view_kwargs(
            **{
                "model": self.model,
                "permission_policy": self.permission_policy,
                "index_url_name": self.get_url_name("index"),
                "index_results_url_name": self.get_url_name("index_results"),
                "history_url_name": self.get_url_name("history"),
                "usage_url_name": self.get_url_name("usage"),
                "add_url_name": self.get_url_name("add"),
                "edit_url_name": self.get_url_name("edit"),
                "delete_url_name": self.get_url_name("delete"),
                "header_icon": self.icon,
                **kwargs,
            }
        )
        if self.copy_view_enabled:
            view_kwargs["copy_url_name"] = self.get_url_name("copy")
        if self.inspect_view_enabled:
            view_kwargs["inspect_url_name"] = self.get_url_name("inspect")
        return view_kwargs

    def get_index_view_kwargs(self, **kwargs):
        view_kwargs = {
            "template_name": self.index_template_name,
            "results_template_name": self.index_results_template_name,
            "list_display": self.list_display,
            "list_filter": self.list_filter,
            "list_export": self.list_export,
            "export_headings": self.export_headings,
            "export_filename": self.export_filename,
            "filterset_class": self.filterset_class,
            "search_fields": self.search_fields,
            "search_backend_name": self.search_backend_name,
            "paginate_by": self.list_per_page,
            **kwargs,
        }
        if self.ordering:
            view_kwargs["default_ordering"] = self.ordering
        return view_kwargs

    def get_add_view_kwargs(self, **kwargs):
        return {
            "panel": self._edit_handler,
            "form_class": self.get_form_class(),
            "template_name": self.create_template_name,
            **kwargs,
        }

    def get_edit_view_kwargs(self, **kwargs):
        return {
            "panel": self._edit_handler,
            "form_class": self.get_form_class(for_update=True),
            "template_name": self.edit_template_name,
            **kwargs,
        }

    def get_delete_view_kwargs(self, **kwargs):
        return {
            "template_name": self.delete_template_name,
            **kwargs,
        }

    def get_history_view_kwargs(self, **kwargs):
        return {
            "template_name": self.history_template_name,
            "history_results_url_name": self.get_url_name("history_results"),
            "header_icon": "history",
            **kwargs,
        }

    def get_usage_view_kwargs(self, **kwargs):
        return {
            "template_name": self.get_templates(
                "usage", fallback=self.usage_view_class.template_name
            ),
            **kwargs,
        }

    def get_inspect_view_kwargs(self, **kwargs):
        return {
            "template_name": self.inspect_template_name,
            "fields": self.inspect_view_fields,
            "fields_exclude": self.inspect_view_fields_exclude,
            **kwargs,
        }

    def get_copy_view_kwargs(self, **kwargs):
        return self.get_add_view_kwargs(**kwargs)

    @property
    def index_view(self):
        return self.construct_view(
            self.index_view_class, **self.get_index_view_kwargs()
        )

    @property
    def index_results_view(self):
        return self.construct_view(
            self.index_view_class, **self.get_index_view_kwargs(), results_only=True
        )

    @property
    def add_view(self):
        return self.construct_view(self.add_view_class, **self.get_add_view_kwargs())

    @property
    def edit_view(self):
        return self.construct_view(self.edit_view_class, **self.get_edit_view_kwargs())

    @property
    def delete_view(self):
        return self.construct_view(
            self.delete_view_class, **self.get_delete_view_kwargs()
        )

    @property
    def history_view(self):
        return self.construct_view(
            self.history_view_class, **self.get_history_view_kwargs()
        )

    @property
    def history_results_view(self):
        return self.construct_view(
            self.history_view_class, **self.get_history_view_kwargs(), results_only=True
        )

    @property
    def usage_view(self):
        return self.construct_view(
            self.usage_view_class, **self.get_usage_view_kwargs()
        )

    @property
    def inspect_view(self):
        return self.construct_view(
            self.inspect_view_class, **self.get_inspect_view_kwargs()
        )

    @property
    def copy_view(self):
        return self.construct_view(self.copy_view_class, **self.get_copy_view_kwargs())

    def get_templates(self, name="index", fallback=""):
        """
        Utility function that provides a list of templates to try for a given
        view, when the template isn't overridden by one of the template
        attributes on the class.
        """
        if not self.template_prefix:
            return [fallback]
        templates = [
            f"{self.template_prefix}{self.app_label}/{self.model_name}/{name}.html",
            f"{self.template_prefix}{self.app_label}/{name}.html",
            f"{self.template_prefix}{name}.html",
        ]
        if fallback:
            templates.append(fallback)
        return templates

    @cached_property
    def index_template_name(self):
        """
        A template to be used when rendering ``index_view``.

        Default: if :attr:`template_prefix` is specified, an ``index.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``index_view_class.template_name`` will be used.
        """
        return self.get_templates(
            "index",
            fallback=self.index_view_class.template_name,
        )

    @cached_property
    def index_results_template_name(self):
        """
        A template to be used when rendering ``index_results_view``.

        Default: if :attr:`template_prefix` is specified, a ``index_results.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``index_view_class.results_template_name`` will be used.
        """
        return self.get_templates(
            "index_results",
            fallback=self.index_view_class.results_template_name,
        )

    @cached_property
    def create_template_name(self):
        """
        A template to be used when rendering ``add_view``.

        Default: if :attr:`template_prefix` is specified, a ``create.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``add_view_class.template_name`` will be used.
        """
        return self.get_templates(
            "create",
            fallback=self.add_view_class.template_name,
        )

    @cached_property
    def edit_template_name(self):
        """
        A template to be used when rendering ``edit_view``.

        Default: if :attr:`template_prefix` is specified, an ``edit.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``edit_view_class.template_name`` will be used.
        """
        return self.get_templates(
            "edit",
            fallback=self.edit_view_class.template_name,
        )

    @cached_property
    def delete_template_name(self):
        """
        A template to be used when rendering ``delete_view``.

        Default: if :attr:`template_prefix` is specified, a ``confirm_delete.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``delete_view_class.template_name`` will be used.
        """
        return self.get_templates(
            "confirm_delete",
            fallback=self.delete_view_class.template_name,
        )

    @cached_property
    def history_template_name(self):
        """
        A template to be used when rendering ``history_view``.

        Default: if :attr:`template_prefix` is specified, a ``history.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``history_view_class.template_name`` will be used.
        """
        return self.get_templates(
            "history",
            fallback=self.history_view_class.template_name,
        )

    @cached_property
    def inspect_template_name(self):
        """
        A template to be used when rendering ``inspect_view``.

        Default: if :attr:`template_prefix` is specified, an ``inspect.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``inspect_view_class.template_name`` will be used.
        """
        return self.get_templates(
            "inspect",
            fallback=self.inspect_view_class.template_name,
        )

    @cached_property
    def list_display(self):
        """
        A list or tuple, where each item is either:

        - The name of a field on the model;
        - The name of a callable or property on the model that accepts a single
          parameter for the model instance; or
        - An instance of the ``wagtail.admin.ui.tables.Column`` class.

        If the name refers to a database field, the ability to sort the listing
        by the database column will be offered and the field's verbose name
        will be used as the column header.

        If the name refers to a callable or property, an ``admin_order_field``
        attribute can be defined on it to point to the database column for
        sorting. A ``short_description`` attribute can also be defined on the
        callable or property to be used as the column header.

        This list will be passed to the ``list_display`` attribute of the index
        view. If left unset, the ``list_display`` attribute of the index view
        will be used instead, which by default is defined as
        ``["__str__", wagtail.admin.ui.tables.LocaleColumn(), wagtail.admin.ui.tables.UpdatedAtColumn()]``.

        Note that the ``LocaleColumn`` is only included if the model is translatable.
        """
        return self.UNDEFINED

    @cached_property
    def list_filter(self):
        """
        A list or tuple, where each item is the name of model fields of type
        ``BooleanField``, ``CharField``, ``DateField``, ``DateTimeField``,
        ``IntegerField`` or ``ForeignKey``.
        Alternatively, it can also be a dictionary that maps a field name to a
        list of lookup expressions.
        This will be passed as django-filter's ``FilterSet.Meta.fields``
        attribute. See
        `its documentation <https://django-filter.readthedocs.io/en/stable/guide/usage.html#generating-filters-with-meta-fields>`_
        for more details.
        If ``filterset_class`` is set, this attribute will be ignored.
        """
        return self.index_view_class.list_filter

    @cached_property
    def filterset_class(self):
        """
        A subclass of ``wagtail.admin.filters.WagtailFilterSet``, which is a
        subclass of `django_filters.FilterSet <https://django-filter.readthedocs.io/en/stable/ref/filterset.html>`_.
        This will be passed to the ``filterset_class`` attribute of the index view.
        """
        return self.UNDEFINED

    @cached_property
    def search_fields(self):
        """
        The fields to use for the search in the index view.
        If set to ``None`` and :attr:`search_backend_name` is set to use a Wagtail search backend,
        the ``search_fields`` attribute of the model will be used instead.
        """
        return self.UNDEFINED

    @cached_property
    def search_backend_name(self):
        """
        The name of the Wagtail search backend to use for the search in the index view.
        If set to a falsy value, the search will fall back to use Django's QuerySet API.
        """
        return self.index_view_class.search_backend_name

    @cached_property
    def list_export(self):
        """
        A list or tuple, where each item is the name of a field, an attribute,
        or a single-argument callable on the model to be exported.
        """
        return self.index_view_class.list_export

    @cached_property
    def export_headings(self):
        """
        A dictionary of export column heading overrides in the format
        ``{field_name: heading}``.
        """
        return self.index_view_class.export_headings

    @cached_property
    def export_filename(self):
        """
        The base file name for the exported listing, without extensions.
        If unset, the model's :attr:`~django.db.models.Options.db_table` will be
        used instead.
        """
        return self.model._meta.db_table

    @cached_property
    def menu_label(self):
        return capfirst(self.model_opts.verbose_name_plural)

    @cached_property
    def menu_item_class(self):
        from wagtail.admin.menu import MenuItem

        def is_shown(_self, request):
            return self.permission_policy.user_has_any_permission(
                request.user, self.index_view_class.any_permission_required
            )

        return type(
            f"{self.model.__name__}MenuItem",
            (MenuItem,),
            {"is_shown": is_shown},
        )

    def formfield_for_dbfield(self, db_field, **kwargs):
        return db_field.formfield(**kwargs)

    def get_form_class(self, for_update=False):
        """
        Returns the form class to use for the create / edit forms.
        """
        # If an edit handler is defined, use it to construct the form class.
        if self._edit_handler:
            return self._edit_handler.get_form_class()

        # Otherwise, use Django's modelform_factory.
        fields = self.get_form_fields()
        exclude = self.get_exclude_form_fields()

        if fields is None and exclude is None:
            raise ImproperlyConfigured(
                "Subclasses of ModelViewSet must specify 'get_form_class', 'form_fields' "
                "or 'exclude_form_fields'."
            )

        return modelform_factory(
            self.model,
            formfield_callback=self.formfield_for_dbfield,
            fields=fields,
            exclude=exclude,
        )

    def get_form_fields(self):
        """
        Returns a list or tuple of field names to be included in the create / edit forms.
        """
        return getattr(self, "form_fields", None)

    def get_exclude_form_fields(self):
        """
        Returns a list or tuple of field names to be excluded from the create / edit forms.
        """
        return getattr(self, "exclude_form_fields", None)

    def get_edit_handler(self):
        """
        Returns the appropriate edit handler for this ``ModelViewSet`` class.
        It can be defined either on the model itself or on the ``ModelViewSet``,
        as the ``edit_handler`` or ``panels`` properties. If none of these are
        defined, it will return ``None`` and the form will be constructed as
        a Django form using :meth:`get_form_class` (without using
        :ref:`forms_panels_overview`).
        """
        if hasattr(self, "edit_handler"):
            edit_handler = self.edit_handler
        elif hasattr(self, "panels"):
            panels = self.panels
            edit_handler = ObjectList(panels)
        elif hasattr(self.model, "edit_handler"):
            edit_handler = self.model.edit_handler
        elif hasattr(self.model, "panels"):
            panels = self.model.panels
            edit_handler = ObjectList(panels)
        else:
            return None
        return edit_handler.bind_to_model(self.model)

    @cached_property
    def _edit_handler(self):
        """
        An edit handler that has been bound to the model class,
        to be used across views.
        """
        return self.get_edit_handler()

    @property
    def url_finder_class(self):
        return type(
            "_ModelAdminURLFinder",
            (ModelAdminURLFinder,),
            {
                "permission_policy": self.permission_policy,
                "edit_url_name": self.get_url_name("edit"),
            },
        )

    def register_admin_url_finder(self):
        register_admin_url_finder(self.model, self.url_finder_class)

    def register_reference_index(self):
        if self.add_to_reference_index:
            ReferenceIndex.register_model(self.model)

    def get_permissions_to_register(self):
        """
        Returns a queryset of :class:`~django.contrib.auth.models.Permission`
        objects to be registered with the :ref:`register_permissions` hook. By
        default, it returns all permissions for the model if
        :attr:`inspect_view_enabled` is set to ``True``. Otherwise, the "view"
        permission is excluded.
        """
        content_type = ContentType.objects.get_for_model(self.model)
        permissions = Permission.objects.filter(content_type=content_type)
        # Only register the "view" permission if the inspect view is enabled
        if not self.inspect_view_enabled:
            permissions = permissions.exclude(
                codename=get_permission_codename("view", self.model_opts)
            )
        return permissions

    def register_permissions(self):
        hooks.register("register_permissions", self.get_permissions_to_register)

    def get_urlpatterns(self):
        urlpatterns = [
            path("", self.index_view, name="index"),
            path("results/", self.index_results_view, name="index_results"),
            path("new/", self.add_view, name="add"),
            path("edit/<str:pk>/", self.edit_view, name="edit"),
            path("delete/<str:pk>/", self.delete_view, name="delete"),
            path("history/<str:pk>/", self.history_view, name="history"),
            path(
                "history-results/<str:pk>/",
                self.history_results_view,
                name="history_results",
            ),
            path("usage/<str:pk>/", self.usage_view, name="usage"),
        ]

        if self.inspect_view_enabled:
            urlpatterns.append(
                path("inspect/<str:pk>/", self.inspect_view, name="inspect")
            )

        if self.copy_view_enabled:
            urlpatterns.append(path("copy/<str:pk>/", self.copy_view, name="copy"))

        return urlpatterns

    def on_register(self):
        super().on_register()
        self.register_admin_url_finder()
        self.register_reference_index()
        self.register_permissions()


class ModelViewSetGroup(ViewSetGroup):
    """
    A container for grouping together multiple
    :class:`~wagtail.admin.viewsets.model.ModelViewSet` instances.

    All attributes and methods from
    :class:`~wagtail.admin.viewsets.base.ViewSetGroup` are available.
    """

    def get_app_label_from_subitems(self):
        for instance in self.registerables:
            if app_label := getattr(instance, "app_label", ""):
                return capfirst(app_label)
        return ""

    @cached_property
    def menu_label(self):
        return self.get_app_label_from_subitems()



from abc import ABC, abstractmethod
from wagtail.blocks.migrations.utils import formatted_list_child_generator
from django.utils.deconstruct import deconstructible


class BaseBlockOperation(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def apply(self, block_value):
        pass

    @property
    @abstractmethod
    def operation_name_fragment(self):
        pass


@deconstructible
class RenameStreamChildrenOperation(BaseBlockOperation):
    """Renames all StreamBlock children of the given type

    Note:
        The `block_path_str` when using this operation should point to the parent StreamBlock
        which contains the blocks to be renamed, not the block being renamed.

    Attributes:
        old_name (str): name of the child block type to be renamed
        new_name (str): new name to rename to
    """

    def __init__(self, old_name, new_name):
        super().__init__()
        self.old_name = old_name
        self.new_name = new_name

    def apply(self, block_value):
        mapped_block_value = []
        for child_block in block_value:
            if child_block["type"] == self.old_name:
                mapped_block_value.append({**child_block, "type": self.new_name})
            else:
                mapped_block_value.append(child_block)
        return mapped_block_value

    @property
    def operation_name_fragment(self):
        return f"rename_{self.old_name}_to_{self.new_name}"


@deconstructible
class RenameStructChildrenOperation(BaseBlockOperation):
    """Renames all StructBlock children of the given type

    Note:
        The `block_path_str` when using this operation should point to the parent StructBlock
        which contains the blocks to be renamed, not the block being renamed.

    Attributes:
        old_name (str): name of the child block type to be renamed
        new_name (str): new name to rename to
    """

    def __init__(self, old_name, new_name):
        super().__init__()
        self.old_name = old_name
        self.new_name = new_name

    def apply(self, block_value):
        mapped_block_value = {}
        for child_key, child_value in block_value.items():
            if child_key == self.old_name:
                mapped_block_value[self.new_name] = child_value
            else:
                mapped_block_value[child_key] = child_value
        return mapped_block_value

    @property
    def operation_name_fragment(self):
        return f"rename_{self.old_name}_to_{self.new_name}"


@deconstructible
class RemoveStreamChildrenOperation(BaseBlockOperation):
    """Removes all StreamBlock children of the given type

    Note:
        The `block_path_str` when using this operation should point to the parent StreamBlock
        which contains the blocks to be removed, not the block being removed.

    Attributes:
        name (str): name of the child block type to be removed
    """

    def __init__(self, name):
        super().__init__()
        self.name = name

    def apply(self, block_value):
        return [
            child_block
            for child_block in block_value
            if child_block["type"] != self.name
        ]

    @property
    def operation_name_fragment(self):
        return f"remove_{self.name}"


@deconstructible
class RemoveStructChildrenOperation(BaseBlockOperation):
    """Removes all StructBlock children of the given type

    Note:
        The `block_path_str` when using this operation should point to the parent StructBlock
        which contains the blocks to be removed, not the block being removed.

    Attributes:
        name (str): name of the child block type to be removed
    """

    def __init__(self, name):
        super().__init__()
        self.name = name

    def apply(self, block_value):
        return {
            child_key: child_value
            for child_key, child_value in block_value.items()
            if child_key != self.name
        }

    @property
    def operation_name_fragment(self):
        return f"remove_{self.name}"


class StreamChildrenToListBlockOperation(BaseBlockOperation):
    """Combines StreamBlock children of the given type into a new ListBlock

    Note:
        The `block_path_str` when using this operation should point to the parent StreamBlock
        which contains the blocks to be combined, not the child block itself.

    Attributes:
        block_name (str): name of the child block type to be combined
        list_block_name (str): name of the new ListBlock type
    """

    def __init__(self, block_name, list_block_name):
        super().__init__()
        self.block_name = block_name
        self.list_block_name = list_block_name

    def apply(self, block_value):
        candidate_blocks = []
        mapped_block_value = []
        for child_block in block_value:
            if child_block["type"] == self.block_name:
                candidate_blocks.append(child_block)
            else:
                mapped_block_value.append(child_block)

        list_items = self.map_temp_blocks_to_list_items(candidate_blocks)

        if list_items:
            new_list_block = {"type": self.list_block_name, "value": list_items}
            mapped_block_value.append(new_list_block)

        return mapped_block_value

    def map_temp_blocks_to_list_items(self, blocks):
        list_items = []
        for block in blocks:
            list_items.append({**block, "type": "item"})
        return list_items

    @property
    def operation_name_fragment(self):
        return f"{self.block_name}_to_list_block_{self.list_block_name}"


class StreamChildrenToStreamBlockOperation(BaseBlockOperation):
    """Combines StreamBlock children of the given types into a new StreamBlock

    Note:
        The `block_path_str` when using this operation should point to the parent StreamBlock
        which contains the blocks to be combined, not the child block itself.

    Attributes:
        block_names (:obj:`list` of :obj:`str`): names of the child block types to be combined
        stream_block_name (str): name of the new StreamBlock type
    """

    def __init__(self, block_names, stream_block_name):
        super().__init__()
        self.block_names = block_names
        self.stream_block_name = stream_block_name

    def apply(self, block_value):
        mapped_block_value = []
        stream_value = []

        for child_block in block_value:
            if child_block["type"] in self.block_names:
                stream_value.append(child_block)
            else:
                mapped_block_value.append(child_block)

        if stream_value:
            new_stream_block = {"type": self.stream_block_name, "value": stream_value}
            mapped_block_value.append(new_stream_block)

        return mapped_block_value

    @property
    def operation_name_fragment(self):
        return "{}_to_stream_block".format("_".join(self.block_names))


class AlterBlockValueOperation(BaseBlockOperation):
    """Alters the value of each block to the given value

    Attributes:
        new_value : new value to change to
    """

    def __init__(self, new_value):
        super().__init__()
        self.new_value = new_value

    def apply(self, block_value):
        return self.new_value

    @property
    def operation_name_fragment(self):
        return "alter_block_value"


class StreamChildrenToStructBlockOperation(BaseBlockOperation):
    """Move each StreamBlock child of the given type inside a new StructBlock

    A new StructBlock will be created as a child of the parent StreamBlock for each child block of
    the given type, and then that child block will be moved from the parent StreamBlocks children
    inside the new StructBlock as a child of that StructBlock.

    Example:
        Consider the following StreamField definition::

            mystream = StreamField([("char1", CharBlock()) ...], ...)

        Then the stream data would look like the following::

            [
                ...
                { "type": "char1", "value": "Value1", ... },
                { "type": "char1", "value": "Value2", ... },
                ...
            ]

        And if we define the operation like this::

            StreamChildrenToStructBlockOperation("char1", "struct1")

        Our altered stream data would look like this::

            [
                ...
                { "type": "struct1", "value": { "char1": "Value1" } },
                { "type": "struct1", "value": { "char1": "Value2" } },
                ...
            ]

    Note:
        The `block_path_str` when using this operation should point to the parent StreamBlock
        which contains the blocks to be combined, not the child block itself.

    Note:
        Block ids are not preserved here since the new blocks are structurally different than the
        previous blocks.

    Attributes:
        block_names (str): names of the child block types to be combined
        struct_block_name (str): name of the new StructBlock type
    """

    def __init__(self, block_name, struct_block_name):
        super().__init__()
        self.block_name = block_name
        self.struct_block_name = struct_block_name

    def apply(self, block_value):
        mapped_block_value = []
        for child_block in block_value:
            if child_block["type"] == self.block_name:
                mapped_block_value.append(
                    {
                        **child_block,
                        "type": self.struct_block_name,
                        "value": {self.block_name: child_block["value"]},
                    }
                )
            else:
                mapped_block_value.append(child_block)
        return mapped_block_value

    @property
    def operation_name_fragment(self):
        return f"{self.block_name}_to_struct_block_{self.struct_block_name}"


class ListChildrenToStructBlockOperation(BaseBlockOperation):
    def __init__(self, block_name):
        super().__init__()
        self.block_name = block_name

    def apply(self, block_value):
        mapped_block_value = []

        # In case there is data from the old list format (wagtail < 2.16), we use the generator
        # to convert them into the new list format
        for child_block in formatted_list_child_generator(block_value):
            mapped_block_value.append(
                {**child_block, "value": {self.block_name: child_block["value"]}}
            )
        return mapped_block_value

    @property
    def operation_name_fragment(self):
        return f"list_block_items_to_{self.block_name}"
